import json
from typing import Annotated, Any, ClassVar, Literal, Optional, Union

import fsspec
import numpy as np
from loguru import logger
from pydantic import (
    AfterValidator,
    BaseModel,
    ConfigDict,
    HttpUrl,
    constr,
    model_validator,
)
from pydantic.alias_generators import to_camel
from typing_extensions import TypeAlias

SplitIndicesType: TypeAlias = list[int]
"""
A split is defined by a sequence of integers.
"""

SplitType: TypeAlias = tuple[SplitIndicesType, Union[SplitIndicesType, dict[str, SplitIndicesType]]]
"""
A split is a pair of which the first item is always assumed to be the train set.
The second item can either be a single test set or a dictionary with multiple, named test sets.
"""

PredictionsType: TypeAlias = Union[np.ndarray, dict[str, Union[np.ndarray, dict[str, np.ndarray]]]]
"""
A prediction is one of three things:

- A single array (single-task, single test set)
- A dictionary of arrays (single-task, multiple test sets) 
- A dictionary of dictionaries of arrays (multi-task, multiple test sets)
"""

DatapointPartType = Union[Any, tuple[Any], dict[str, Any]]
DatapointType: TypeAlias = tuple[DatapointPartType, DatapointPartType]
"""
A datapoint has:

- A single input or multiple inputs (either as dict or tuple)
- No target, a single target or a multiple targets (either as dict or tuple)
"""

DataFormat: TypeAlias = Literal["dict", "tuple"]
"""
The target formats that are supported by the `Subset` class. 
"""

SlugStringType: TypeAlias = constr(pattern="^[a-z0-9-]+$", min_length=4, max_length=64)
"""
A URL-compatible string that can serve as slug on the hub.
"""

SlugCompatibleStringType: TypeAlias = constr(pattern="^[A-Za-z0-9_-]+$", min_length=4, max_length=64)
"""
A URL-compatible string that can be turned into a slug by the hub.

Can only use alpha-numeric characters, underscores and dashes. 
The string must be at least 4 and at most 64 characters long.
"""


HubUser: TypeAlias = SlugCompatibleStringType
"""
A user on the Polaris Hub is identified by a username, 
which is a [`SlugCompatibleStringType`][polaris.utils.types.SlugCompatibleStringType].
"""

HttpUrlString: TypeAlias = Annotated[HttpUrl, AfterValidator(str)]
"""
A validated URL that will be turned into a string.
This is useful for interactions with httpx and authlib, who have their own URL types.
"""

DirectionType: TypeAlias = Literal["min", "max"]
"""
The direction of any variable to be sorted.
This can be used to sort the metric score, indicate the optmization direction of endpoint.
"""

AccessType: TypeAlias = Literal["public", "private"]
"""
Type to specify access to a dataset, benchmark or result in the Hub.
"""


class HubOwner(BaseModel):
    """An owner of an artifact on the Polaris Hub

    The slug is most important as it is the user-facing part of this data model.
    The externalId and type are added to be consistent with the model returned by the Polaris Hub .
    """

    slug: SlugStringType
    external_id: Optional[str] = None
    type: Optional[Literal["user", "organization"]] = None

    # Pydantic config
    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True)

    def __str__(self):
        return self.slug


class License(BaseModel):
    """An artifact license.

    Attributes:
        id: The license ID. Either from [SPDX](https://spdx.org/licenses/) or custom.
        reference: A reference to the license text. If the ID is found in SPDX, this is automatically set.
            Else it is required to manually specify this.
    """

    SPDX_LICENSE_DATA_PATH: ClassVar[
        str
    ] = "https://raw.githubusercontent.com/spdx/license-list-data/main/json/licenses.json"

    id: str
    reference: Optional[HttpUrlString] = None

    @model_validator(mode="after")  # type: ignore
    @classmethod
    def _validate_license_id(cls, m: "License"):
        """
        If a license ID exists in the SPDX database, we use the reference from there.
        Otherwise, it is up to the user to specify the license.
        """

        # Load the ground truth references
        with fsspec.open(cls.SPDX_LICENSE_DATA_PATH) as f:
            data = json.load(f)
        data = {license["licenseId"]: license for license in data["licenses"]}

        if m.id in data:
            if m.reference is not None and m.reference != data[m.id]["reference"]:
                logger.warning(f"Found license ID {m.id} in SPDX, using the associated reference.")
            m.reference = data[m.id]["reference"]

        if m.id not in data and m.reference is None:
            raise ValueError(
                f"License with ID {m.id} not found in SPDX. "
                "It is required to then also specify the name and reference."
            )
        return m



from enum import Enum
from typing import Callable

import numpy as np
from pydantic import BaseModel, Field
from scipy import stats
from sklearn.metrics import (
    accuracy_score,
    average_precision_score,
    cohen_kappa_score,
    explained_variance_score,
    f1_score,
    matthews_corrcoef,
    mean_absolute_error,
    mean_squared_error,
    r2_score,
    roc_auc_score,
)

from polaris.utils.types import DirectionType


def pearsonr(y_true: np.ndarray, y_pred: np.ndarray):
    """Calculate a pearson r correlation"""
    return stats.pearsonr(y_true, y_pred).statistic


def spearman(y_true: np.ndarray, y_pred: np.ndarray):
    """Calculate a Spearman correlation"""
    return stats.spearmanr(y_true, y_pred).statistic


class MetricInfo(BaseModel):
    """
    Metric metadata

    Attributes:
        fn: The callable that actually computes the metric.
        is_multitask: Whether the metric expects a single set of predictions or a dict of predictions.
        kwargs: Additional parameters required for the metric.
        direction:  The direction for ranking of the metric,  "max" for maximization and "min" for minimization.
    """

    fn: Callable
    is_multitask: bool = False
    kwargs: dict = Field(default_factory=dict)
    direction: DirectionType


class Metric(Enum):
    """
    A metric within the Polaris ecosystem is uniquely identified by its name
    and is associated with additional metadata in a `MetricInfo` instance.

    Implemented as an enum.
    """

    # TODO (cwognum):
    #  - Any preprocessing needed? For example changing the shape / dtype? Converting from torch tensors or lists?

    # regression
    mean_absolute_error = MetricInfo(fn=mean_absolute_error, direction="min")
    mean_squared_error = MetricInfo(fn=mean_squared_error, direction="min")
    r2 = MetricInfo(fn=r2_score, direction="max")
    pearsonr = MetricInfo(fn=pearsonr, direction="max")
    spearmanr = MetricInfo(fn=spearman, direction="max")
    explained_var = MetricInfo(fn=explained_variance_score, direction="max")

    # classification
    accuracy = MetricInfo(fn=accuracy_score, direction="max")
    f1 = MetricInfo(fn=f1_score, kwargs={"average": "binary"}, direction="max")
    f1_macro = MetricInfo(fn=f1_score, kwargs={"average": "macro"}, direction="max")
    f1_micro = MetricInfo(fn=f1_score, kwargs={"average": "micro"}, direction="max")
    roc_auc = MetricInfo(fn=roc_auc_score, direction="max")
    pr_auc = MetricInfo(fn=average_precision_score, direction="max")
    mcc = MetricInfo(fn=matthews_corrcoef, direction="max")
    cohen_kappa = MetricInfo(fn=cohen_kappa_score, direction="max")
    # TODO: adding metrics for multiclass tasks

    @property
    def fn(self) -> Callable:
        """The callable that actually computes the metric"""
        return self.value.fn

    @property
    def is_multitask(self) -> bool:
        """Whether the metric expects a single set of predictions or a dict of predictions."""
        return self.value.is_multitask

    def score(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        """Endpoint for computing the metric.

        For convenience, calling a `Metric` will result in this method being called.

        ```python
        metric = Metric.mean_absolute_error
        assert metric.score(y_true=first, y_pred=second) == metric(y_true=first, y_pred=second)
        ```
        """
        return self.fn(y_true, y_pred, **self.value.kwargs)

    def __call__(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        """For convenience, make metrics callable"""
        return self.score(y_true, y_pred)



from typing import List, Literal, Optional, Sequence, Union

import numpy as np

from polaris.dataset import Dataset
from polaris.utils.context import tmp_attribute_change
from polaris.utils.errors import TestAccessError
from polaris.utils.types import DataFormat, DatapointType


class Subset:
    """The `Subset` class provides easy access to a single partition of a split dataset.

    This should be the starting point for any framework-specific (e.g. PyTorch, Tensorflow) data-loader implementation.
    How the data is loaded in Polaris can be non-trivial, so this class is provided to abstract away the details.
    To easily build framework-specific data-loaders, a `Subset` supports various styles of accessing the data:

    1. **In memory**: Loads the entire dataset in memory and returns a single array with all datapoints,
        this style is accessible through the `subset.targets` and `subset.inputs` properties.
    2. **List**: Index the subset like a list, this style is accessible through the `subset[idx]` syntax.
    3. **Iterator**: Iterate over the subset, this style is accessible through the `iter(subset)` syntax.

    Examples:
        The different styles of accessing the data:

        ```python
        import polaris as po

        benchmark = po.load_benchmark(...)
        train, test = benchmark.get_train_test_split()

        # Load the entire dataset in memory, useful for e.g. scikit-learn.
        X = train.inputs
        y = train.targets

        # Access a single datapoint as with a list, useful for e.g. PyTorch.
        x, y = train[0]

        # Iterate over the dataset, useful for very large datasets.
        for x, y in train:
            ...
        ```

    Raises:
        TestAccessError: When trying to access the targets of the test set (specified by the `hide_targets` attribute).
    """

    _SUPPORTED_FORMATS = ["dict", "tuple"]

    def __init__(
        self,
        dataset: Dataset,
        indices: List[Union[int, Sequence[int]]],
        input_cols: Union[List[str], str],
        target_cols: Union[List[str], str],
        input_format: DataFormat = "dict",
        target_format: DataFormat = "tuple",
        hide_targets: bool = False,
    ):
        self.dataset = dataset
        self.indices = indices
        self.target_cols = target_cols if isinstance(target_cols, list) else [target_cols]
        self.input_cols = input_cols if isinstance(input_cols, list) else [input_cols]

        # Validate the output format
        if input_format not in self._SUPPORTED_FORMATS:
            raise ValueError(
                f"Unsupported output format {input_format}. Choose from {self._SUPPORTED_FORMATS}"
            )
        if target_format not in self._SUPPORTED_FORMATS:
            raise ValueError(
                f"Unsupported output format {target_format}. Choose from {self._SUPPORTED_FORMATS}"
            )
        self._input_format = input_format
        self._target_format = target_format

        # For the iterator implementation
        self._pointer = 0

        # This is a protected attribute to make explicit it should not be changed once set.
        self._hide_targets = hide_targets

    @property
    def is_multi_task(self):
        return len(self.target_cols) > 1

    @property
    def is_multi_input(self):
        return len(self.input_cols) > 1

    @property
    def inputs(self):
        """
        Scikit-learn style access to the inputs.
        If the dataset is multi-input, this will return a dict of arrays.
        """
        return self.as_array("x")

    @property
    def targets(self):
        """
        Scikit-learn style access to the targets.
        If the dataset is multi-target, this will return a dict of arrays.
        """
        return self.as_array("y")

    @staticmethod
    def _convert(data: dict, order: List[str], fmt: str):
        """Converts from the default dict format to the specified format"""
        if len(data) == 1:
            data = list(data.values())[0]
        elif fmt == "tuple":
            data = tuple(data[k] for k in order)
        return data

    def _extract(
        self,
        data: DatapointType,
        data_type: Union[Literal["x"], Literal["y"], Literal["xy"]],
        key: Optional[str] = None,
    ):
        """Helper function to extract data from the return format of this class"""
        if self._hide_targets:
            return data
        x, y = data
        ret = x if data_type == "x" else y
        if not isinstance(ret, dict) or key is None:
            return ret
        return ret[key]

    def as_array(self, data_type: Union[Literal["x"], Literal["y"], Literal["xy"]]):
        """
        Scikit-learn style access to the targets and inputs.
        If the dataset is multi-target, this will return a dict of arrays.
        """

        if data_type == "xy":
            return self.as_array("x"), self.as_array("y")

        if data_type == "y" and self._hide_targets:
            raise TestAccessError("Within Polaris, you should not need to access the targets of the test set")

        if not self.is_multi_task:
            return np.array([self._extract(ret, data_type) for ret in self])

        out = {}
        columns = self.input_cols if data_type == "x" else self.target_cols

        # Temporarily change the target format for easier conversion
        with tmp_attribute_change(self, "_target_format", "dict"):
            with tmp_attribute_change(self, "_input_format", "dict"):
                for k in columns:
                    out[k] = np.array([self._extract(ret, data_type, k) for ret in self])

        return self._convert(out, self.target_cols, self._target_format)

    def __len__(self):
        return len(self.indices)

    def __getitem__(self, item) -> DatapointType:
        """
        This method always returns an (X, y) tuple

        Some special cases:

        1. It supports multi-input datasets, in which case X is a dict with multiple values.
        2. It supports multi-task datasets, in which case y is a dict with multiple values.
        3. In case a dataset has a pointer column (i.e. a path to an external file with the actual data),
            this method also loads that data to memory.
        """

        idx = self.indices[item]

        # Get the row from the base table
        row = self.dataset.table.iloc[idx]

        # Load the input modalities
        ins = {col: self.dataset.get_data(row.name, col) for col in self.input_cols}
        ins = self._convert(ins, self.input_cols, self._input_format)

        if self._hide_targets:
            # If we are not allowed to access the targets, we return the inputs only.
            # This is useful to make accidental access to the test set less likely.
            return ins

        # Retrieve the targets
        outs = {col: self.dataset.get_data(row.name, col) for col in self.target_cols}
        outs = self._convert(outs, self.target_cols, self._target_format)

        return ins, outs

    def __iter__(self):
        """Iterator implementation"""
        self._pointer = 0
        return self

    def __next__(self):
        """
        Allows for iterator-style access to the dataset, e.g. useful when datasets get large.
        Remember that pointer columns are not loaded into memory yet.
        """
        if self._pointer >= len(self):
            raise StopIteration

        item = self[self._pointer]
        self._pointer += 1
        return item



from contextlib import contextmanager


@contextmanager
def tmp_attribute_change(obj, attribute, value):
    """Temporarily set and reset an attribute of an object."""
    original_value = getattr(obj, attribute)
    try:
        setattr(obj, attribute, value)
        yield obj
    finally:
        setattr(obj, attribute, original_value)



import datamol as dm
import numpy as np
import pytest
import zarr

from polaris.benchmark import (
    MultiTaskBenchmarkSpecification,
    SingleTaskBenchmarkSpecification,
)
from polaris.dataset import ColumnAnnotation, Dataset
from polaris.utils import fs
from polaris.utils.types import HubOwner, License


def _get_zarr_archive(tmp_path, datapoint_per_array: bool):
    tmp_path = fs.join(str(tmp_path), "data.zarr")
    root = zarr.open_group(tmp_path, mode="w")
    group_a = root.create_group("A/")
    group_b = root.create_group("B/")

    def _populate_group(group):
        if datapoint_per_array:
            for i in range(100):
                group.array(i, data=np.random.random((2048,)))
        else:
            group.array("data", data=np.random.random((100, 2048)))

    _populate_group(group_a)
    _populate_group(group_b)
    return tmp_path


@pytest.fixture(scope="module")
def test_data():
    data = dm.data.freesolv()[:100]
    # set an abitrary threshold for testing purpose.
    data["CLASS_expt"] = data["expt"].gt(0).astype(int).values
    data["CLASS_calc"] = data["calc"].gt(0).astype(int).values
    return data


@pytest.fixture(scope="module")
def test_org_owner():
    return HubOwner(organizationId="test-organization", slug="test-organization")


@pytest.fixture(scope="module")
def test_user_owner():
    return HubOwner(userId="test-user", slug="test-user")


@pytest.fixture(scope="module")
def test_dataset(test_data, test_org_owner):
    return Dataset(
        table=test_data,
        name="test-dataset",
        source="https://www.example.com",
        annotations={"expt": ColumnAnnotation(is_pointer=False, user_attributes={"unit": "kcal/mol"})},
        tags=["tagA", "tagB"],
        user_attributes={"attributeA": "valueA", "attributeB": "valueB"},
        owner=test_org_owner,
        license=License(id="MIT"),
    )


@pytest.fixture(scope="function")
def test_zarr_archive_multiple_arrays(tmp_path):
    return _get_zarr_archive(tmp_path, datapoint_per_array=True)


@pytest.fixture(scope="function")
def test_zarr_archive_single_array(tmp_path):
    return _get_zarr_archive(tmp_path, datapoint_per_array=False)


@pytest.fixture(scope="module")
def test_single_task_benchmark(test_dataset):
    train_indices = list(range(90))
    test_indices = list(range(90, 100))
    return SingleTaskBenchmarkSpecification(
        name="single-task-benchmark",
        dataset=test_dataset,
        metrics=[
            "mean_absolute_error",
            "mean_squared_error",
            "r2",
            "spearmanr",
            "pearsonr",
            "explained_var",
        ],
        main_metric="mean_absolute_error",
        split=(train_indices, test_indices),
        target_cols="expt",
        input_cols="smiles",
    )


@pytest.fixture(scope="module")
def test_single_task_benchmark_clf(test_dataset):
    train_indices = list(range(90))
    test_indices = list(range(90, 100))
    return SingleTaskBenchmarkSpecification(
        name="single-task-benchmark",
        dataset=test_dataset,
        main_metric="accuracy",
        metrics=["accuracy", "f1", "roc_auc", "pr_auc", "mcc", "cohen_kappa"],
        split=(train_indices, test_indices),
        target_cols="CLASS_expt",
        input_cols="smiles",
    )


@pytest.fixture(scope="module")
def test_single_task_benchmark_multiple_test_sets(test_dataset):
    train_indices = list(range(90))
    test_indices = {"test_1": list(range(90, 95)), "test_2": list(range(95, 100))}
    return SingleTaskBenchmarkSpecification(
        name="single-task-benchmark",
        dataset=test_dataset,
        metrics=[
            "mean_absolute_error",
            "mean_squared_error",
            "r2",
            "spearmanr",
            "pearsonr",
            "explained_var",
        ],
        main_metric="r2",
        split=(train_indices, test_indices),
        target_cols="expt",
        input_cols="smiles",
    )


@pytest.fixture(scope="module")
def test_multi_task_benchmark(test_dataset):
    # For the sake of simplicity, just use a small set of indices
    train_indices = list(range(90))
    test_indices = list(range(90, 100))
    return MultiTaskBenchmarkSpecification(
        name="multi-task-benchmark",
        dataset=test_dataset,
        main_metric="mean_absolute_error",
        metrics=[
            "mean_absolute_error",
            "mean_squared_error",
            "r2",
            "spearmanr",
            "pearsonr",
            "explained_var",
        ],
        split=(train_indices, test_indices),
        target_cols=["expt", "calc"],
        input_cols="smiles",
    )


@pytest.fixture(scope="module")
def test_multi_task_benchmark_clf(test_dataset):
    # For the sake of simplicity, just use a small set of indices
    train_indices = list(range(90))
    test_indices = list(range(90, 100))
    return MultiTaskBenchmarkSpecification(
        name="multi-task-benchmark",
        dataset=test_dataset,
        main_metric="accuracy",
        metrics=["accuracy", "f1", "roc_auc", "pr_auc", "mcc", "cohen_kappa"],
        split=(train_indices, test_indices),
        target_cols=["CLASS_expt", "CLASS_calc"],
        input_cols="smiles",
    )



import json
import os
import ssl
import webbrowser
from hashlib import md5
from io import BytesIO
from typing import Callable, Optional, Union
from urllib.parse import urljoin

import certifi
import fsspec
import httpx
import pandas as pd
from authlib.common.security import generate_token
from authlib.integrations.base_client.errors import InvalidTokenError, MissingTokenError
from authlib.integrations.httpx_client import OAuth2Client, OAuthError
from authlib.oauth2.client import OAuth2Client as _OAuth2Client
from httpx import HTTPStatusError
from httpx._types import HeaderTypes, TimeoutTypes, URLTypes
from loguru import logger

from polaris.benchmark import (
    BenchmarkSpecification,
    MultiTaskBenchmarkSpecification,
    SingleTaskBenchmarkSpecification,
)
from polaris.dataset import Dataset
from polaris.evaluate import BenchmarkResults
from polaris.hub.settings import PolarisHubSettings
from polaris.utils import fs
from polaris.utils.constants import DEFAULT_CACHE_DIR
from polaris.utils.errors import PolarisHubError, PolarisUnauthorizedError
from polaris.utils.types import AccessType, HubOwner

_HTTPX_SSL_ERROR_CODE = "[SSL: CERTIFICATE_VERIFY_FAILED]"


class PolarisHubClient(OAuth2Client):
    """
    A client for the Polaris Hub API. The Polaris Hub is a central repository of datasets, benchmarks and results.
    Visit it here: [https://polarishub.io/](https://polarishub.io/).

    Bases the [`authlib` client](https://docs.authlib.org/en/latest/client/api.html#authlib.integrations.httpx_client.OAuth2Client),
    which in turns bases the [`httpx` client](https://www.python-httpx.org/advanced/#client-instances).
    See the relevant docs to learn more about how to use these clients outside of the integration with the Polaris Hub.

    Note: Closing the client
        The client should be closed after all requests have been made. For convenience, you can also use the
        client as a context manager to automatically close the client when the context is exited. Note that once
        the client has been closed, it cannot be used anymore.

        ```python
        # Make sure to close the client once finished
        client = PolarisHubClient()
        client.get(...)
        client.close()

        # Or use the client as a context manager
        with PolarisHubClient() as client:
            client.get(...)
        ```

    Info: Async Client
        `authlib` also supports an [async client](https://docs.authlib.org/en/latest/client/httpx.html#async-oauth-2-0).
        Since we don't expect to make multiple requests to the Hub in parallel
        and due to the added complexity stemming from using the Python asyncio API,
        we are sticking to the sync client - at least for now.
    """

    def __init__(
        self,
        env_file: Optional[Union[str, os.PathLike]] = None,
        settings: Optional[PolarisHubSettings] = None,
        cache_auth_token: bool = True,
        **kwargs: dict,
    ):
        """
        Args:
            env_file: Path to a `.env` file containing the settings, used to initialize
                a `PolarisHubSettings` instance. If not provided, the default settings are used.
            settings: A `PolarisHubSettings` instance. If provided, takes precedence over `env_file`.
            cache_auth_token: Whether to cache the auth token to a file.
            **kwargs: Additional keyword arguments passed to the authlib `OAuth2Client` constructor.
        """
        self._user_info = None

        if settings is None:
            settings = PolarisHubSettings(_env_file=env_file)  # type: ignore
        self.settings = settings

        # We cache the auth token by default, but allow the user to disable this.
        self.cache_auth_token = cache_auth_token
        token = kwargs.get("token")
        if fs.exists(self._auth_token_cache_path) and token is None:  # type: ignore
            with fsspec.open(self._auth_token_cache_path, "r") as fd:
                token = json.load(fd)  # type: ignore

        verify = self.settings.ca_bundle
        if verify is None:
            verify = True

        self.code_verifier = generate_token(48)

        super().__init__(
            # OAuth2Client
            client_id=settings.client_id,
            redirect_uri=settings.callback_url,
            scope=settings.scopes,
            token=token,
            token_endpoint=self.settings.token_fetch_url,
            code_challenge_method="S256",
            # httpx.Client
            base_url=settings.api_url,
            verify=verify,
            # Extra
            **kwargs,
        )

    def _load_from_signed_url(self, url: URLTypes, load_fn: Callable, headers: Optional[HeaderTypes] = None):
        """Utility function to load a file from a signed URL"""
        response = self.get(url, auth=None, headers=headers)  # type: ignore
        response.raise_for_status()
        content = BytesIO(response.content)
        return load_fn(content)

    def _base_request_to_hub(self, url: str, method: str, **kwargs):
        """Utility function since most API methods follow the same pattern"""
        response = self.request(url=url, method=method, **kwargs)

        try:
            response.raise_for_status()

        except HTTPStatusError as error:
            # With an internal server error, we are not sure the custom error-handling code on the hub is reached.
            if response.status_code == 500:
                raise

            # If not an internal server error, the hub should always return a JSON response
            # with additional information about the error.
            response = response.json()
            response = json.dumps(response, indent=2, sort_keys=True)
            raise PolarisHubError(
                f"The request to the Polaris Hub failed. See the error message below for more details:\n{response}"
            ) from error

        # Convert the response to json format if the response contains a 'text' body
        try:
            response = response.json()
        except json.JSONDecodeError:
            pass

        return response

    # =========================
    #     Overrides
    # =========================

    @_OAuth2Client.token.setter
    def token(self, value):
        """Override the token setter to additionally save the token to a cache"""
        super(OAuth2Client, type(self)).token.fset(self, value)  # type: ignore

        # We cache afterwards, because the token setter adds fields we need to save (i.e. expires_at).
        if self.cache_auth_token:
            with fsspec.open(self._auth_token_cache_path, "w") as fd:
                json.dump(value, fd)  # type: ignore

    @property
    def _auth_token_cache_path(self) -> str:
        """If self.cache_auth_token is True, this is the location of the cached auth token."""
        path = None
        if self.cache_auth_token:
            fs.mkdir(DEFAULT_CACHE_DIR, exist_ok=True)
            path = fs.join(DEFAULT_CACHE_DIR, "polaris_auth_token.json")
        return path

    def create_authorization_url(self, **kwargs) -> tuple[str, Optional[str]]:
        """Light wrapper to automatically pass in the right URL."""
        return super().create_authorization_url(
            url=self.settings.authorize_url, code_verifier=self.code_verifier, **kwargs
        )

    def fetch_token(self, **kwargs):
        """Light wrapper to automatically pass in the right URL"""
        return super().fetch_token(
            url=self.settings.token_fetch_url, code_verifier=self.code_verifier, **kwargs
        )

    def request(self, method, url, withhold_token=False, auth=httpx.USE_CLIENT_DEFAULT, **kwargs):
        """Wraps the base request method to handle errors"""
        try:
            response = super().request(method, url, withhold_token, auth, **kwargs)
        except httpx.ConnectError as error:
            # NOTE (cwognum): In the stack-trace, the more specific SSLCertVerificationError is raised.
            #   We could use the traceback module to cactch this specific error, but that would be overkill here.
            if _HTTPX_SSL_ERROR_CODE in str(error):
                raise ssl.SSLCertVerificationError(
                    "We could not verify the SSL certificate. "
                    f"Please ensure the installed version ({certifi.__version__}) of the `certifi` package is the latest. "
                    "If you require the usage of a custom CA bundle, you can set the POLARIS_CA_BUNDLE "
                    "environment variable to the path of your CA bundle. For debugging, you can temporarily disable "
                    "SSL verification by setting the POLARIS_CA_BUNDLE environment variable to `false`."
                ) from error
            raise error
        except (MissingTokenError, InvalidTokenError, httpx.HTTPStatusError, OAuthError) as error:
            if isinstance(error, httpx.HTTPStatusError) and error.response.status_code != 401:
                raise
            raise PolarisUnauthorizedError from error
        return response

    # =========================
    #     API Endpoints
    # =========================

    @property
    def user_info(self) -> dict:
        """Get information about the currently logged in user through the OAuth2 User Info flow."""

        # NOTE (cwognum): We override the default `auth` and `headers` argument, since
        #  the defaults trigger a 530 error (Cloudflare) due to the header ordering.
        #  Because of this, we also have to copy some code from the base `request` method to
        #  make auto-refresh a token if needed. For more info, see: https://stackoverflow.com/a/62687390

        try:
            if self.token is None or not self.ensure_active_token(self.token):
                raise PolarisUnauthorizedError
        except OAuthError:
            raise PolarisUnauthorizedError

        if self._user_info is None:
            user_info = self.get(
                self.settings.user_info_url,
                auth=None,  # type: ignore
                headers={"authorization": f"Bearer {self.token['access_token']}"},
            )
            user_info.raise_for_status()
            self._user_info = user_info.json()

        return self._user_info

    @property
    def user_as_owner(self) -> HubOwner:
        """Easily get the currently logged-in user a `HubOwner` instance."""
        return HubOwner(user_id=self.user_info["user_id"], slug=self.user_info["username"])

    def login(self, overwrite: bool = False, auto_open_browser: bool = True):
        """Login to the Polaris Hub using the OAuth2 protocol.

        Warning: Headless authentication
            It is currently not possible to login to the Polaris Hub without a browser.
            See [this Github issue](https://github.com/polaris-hub/polaris/issues/30) for more info.

        Args:
            overwrite: Whether to overwrite the current token if the user is already logged in.
            auto_open_browser: Whether to automatically open the browser to visit the authorization URL.
        """

        # Check if the user is already logged in
        if self.token is not None and not overwrite:
            try:
                info = self.user_info
                logger.info(
                    f"You are already logged in to the Polaris Hub as {info['username']} ({info['email']}). "
                    "Set `overwrite=True` to force re-authentication."
                )
                return
            except PolarisUnauthorizedError:
                pass

        # Step 1: Redirect user to the authorization URL
        authorization_url, _ = self.create_authorization_url()

        if auto_open_browser:
            logger.info(f"Your browser has been opened to visit:\n{authorization_url}\n")
            webbrowser.open_new_tab(authorization_url)
        else:
            logger.info(f"Please visit the following URL:\n{authorization_url}\n")

        # Step 2: After user grants permission, we'll get the authorization code through the callback URL
        authorization_code = input("Please enter the authorization token: ")

        # Step 3: Exchange authorization code for an access token
        self.fetch_token(code=authorization_code, grant_type="authorization_code")

        logger.success(
            f"Successfully authenticated to the Polaris Hub "
            f"as `{self.user_info['username']}` ({self.user_info['email']})! 🎉"
        )

    def list_datasets(self, limit: int = 100, offset: int = 0) -> list[str]:
        """List all available datasets on the Polaris Hub.

        Args:
            limit: The maximum number of datasets to return.
            offset: The offset from which to start returning datasets.

        Returns:
            A list of dataset names in the format `owner/dataset_name`.
        """
        response = self._base_request_to_hub(
            url="/dataset", method="GET", params={"limit": limit, "offset": offset}
        )
        dataset_list = [bm["artifactId"] for bm in response["data"]]
        return dataset_list

    def get_dataset(self, owner: Union[str, HubOwner], name: str) -> Dataset:
        """Load a dataset from the Polaris Hub.

        Args:
            owner: The owner of the dataset. Can be either a user or organization from the Polaris Hub.
            name: The name of the dataset.

        Returns:
            A `Dataset` instance, if it exists.
        """

        response = self._base_request_to_hub(url=f"/dataset/{owner}/{name}", method="GET")
        storage_response = self.get(response["tableContent"]["url"])

        # This should be a 307 redirect with the signed URL
        if storage_response.status_code != 307:
            raise PolarisHubError("Could not get signed URL from Polaris Hub.")

        storage_response = storage_response.json()
        url = storage_response["url"]
        headers = storage_response["headers"]

        response["table"] = self._load_from_signed_url(url=url, headers=headers, load_fn=pd.read_parquet)

        return Dataset(**response)

    def list_benchmarks(self, limit: int = 100, offset: int = 0) -> list[str]:
        """List all available benchmarks on the Polaris Hub.

        Args:
            limit: The maximum number of benchmarks to return.
            offset: The offset from which to start returning benchmarks.

        Returns:
            A list of benchmark names in the format `owner/benchmark_name`.
        """

        # TODO (cwognum): What to do with pagination, i.e. limit and offset?
        response = self._base_request_to_hub(
            url="/benchmark", method="GET", params={"limit": limit, "offset": offset}
        )
        benchmarks_list = [f"{HubOwner(**bm['owner'])}/{bm['name']}" for bm in response["data"]]
        return benchmarks_list

    def get_benchmark(self, owner: Union[str, HubOwner], name: str) -> BenchmarkSpecification:
        """Load a benchmark from the Polaris Hub.

        Args:
            owner: The owner of the benchmark. Can be either a user or organization from the Polaris Hub.
            name: The name of the benchmark.

        Returns:
            A `BenchmarkSpecification` instance, if it exists.
        """

        response = self._base_request_to_hub(url=f"/benchmark/{owner}/{name}", method="GET")

        # TODO (cwognum): Currently, the benchmark endpoints do not return the owner info for the underlying dataset.
        # TODO (jstlaurent): Use the same owner for now, until the benchmark returns a better dataset entity
        response["dataset"] = self.get_dataset(owner, response["dataset"]["name"])

        # TODO (cwognum): As we get more complicated benchmarks, how do we still find the right subclass?
        #  Maybe through structural pattern matching, introduced in Py3.10, or Pydantic's discriminated unions?
        benchmark_cls = (
            SingleTaskBenchmarkSpecification
            if len(response["targetCols"]) == 1
            else MultiTaskBenchmarkSpecification
        )
        return benchmark_cls(**response)

    def upload_results(self, results: BenchmarkResults, access: AccessType = "private"):
        """Upload the results to the Polaris Hub.

        Info: Owner
            The owner of the results will automatically be inferred by the hub from the user making the request.
            Contrary to other artifact types, an organization cannot own a set of results.
            However, you can specify the `BenchmarkResults.contributors` field to share credit with other hub users.

        Note: Required meta-data
            The Polaris client and hub maintain different requirements as to which meta-data is required.
            The requirements by the hub are stricter, so when uploading to the hub you might
            get some errors on missing meta-data. Make sure to fill-in as much of the meta-data as possible
            before uploading.

        Note: Benchmark name and owner
            Importantly, `results.benchmark_name` and `results.benchmark_owner` must be specified
            and match an existing benchmark on the Polaris Hub. If these results were generated by
            `benchmark.evaluate(...)`, this is done automatically.

        Args:
            results: The results to upload.
            access: Grant public or private access to result
        """

        # Get the serialized model data-structure
        result_json = results.model_dump(by_alias=True, exclude_none=True)

        # Make a request to the hub
        response = self._base_request_to_hub(
            url="/result", method="POST", json={"access": access, **result_json}
        )

        # Inform the user about where to find their newly created artifact.
        result_url = urljoin(
            self.settings.hub_url,
            f"benchmarks/{results.benchmark_owner}/{results.benchmark_name}/{response['id']}",
        )
        logger.success(f"Your result has been successfully uploaded to the Hub. View it here: {result_url}")
        return response

    def upload_dataset(
        self, dataset: Dataset, access: AccessType = "private", timeout: TimeoutTypes = (10, 200)
    ):
        """Upload the dataset to the Polaris Hub.

        Info: Owner
            You have to manually specify the owner in the dataset data model. Because the owner could
            be a user or an organization, we cannot automatically infer this from just the logged-in user.

        Note: Required meta-data
            The Polaris client and hub maintain different requirements as to which meta-data is required.
            The requirements by the hub are stricter, so when uploading to the hub you might
            get some errors on missing meta-data. Make sure to fill-in as much of the meta-data as possible
            before uploading.

        Args:
            dataset: The dataset to upload.
            access: Grant public or private access to result
            timeout: Request timeout values. User can modify the value when uploading large dataset as needed.
        """

        # Get the serialized data-model
        # We exclude the table as it handled separately and the cache_dir as it is user-specific
        dataset_json = dataset.model_dump(exclude={"cache_dir", "table"}, exclude_none=True, by_alias=True)

        # Uploading a dataset is a two-step process.
        # 1. Upload the dataset meta data to the hub and prepare the hub to receive the parquet file
        # 2. Upload the parquet file to the hub
        # TODO: Revert step 1 in case step 2 fails - Is this needed? Or should this be taken care of by the hub?

        # Write the parquet file directly to a buffer
        buffer = BytesIO()
        dataset.table.to_parquet(buffer, engine="auto")
        parquet_size = len(buffer.getbuffer())
        parquet_md5 = md5(buffer.getbuffer()).hexdigest()

        # Step 1: Upload meta-data
        # Instead of directly uploading the table, we announce to the hub that we intend to upload one.
        url = f"/dataset/{dataset.artifact_id}"
        response = self._base_request_to_hub(
            url=url,
            method="PUT",
            json={
                "tableContent": {
                    "size": parquet_size,
                    "fileType": "parquet",
                    "md5sum": parquet_md5,
                },
                "access": access,
                **dataset_json,
            },
        )

        # Step 2: Upload the parquet file
        # create an empty PUT request to get the table content URL from cloudflare
        hub_response = self.request(
            url=response["tableContent"]["url"],
            method="PUT",
            headers={
                "Content-type": "application/vnd.apache.parquet",
            },
        )

        if hub_response.status_code == 307:
            # If the hub returns a 307 redirect, we need to follow it to get the signed URL
            hub_response_body = hub_response.json()
            # Upload the data to the cloudflare url
            bucket_response = self.request(
                url=hub_response_body["url"],
                method=hub_response_body["method"],
                headers={"Content-type": "application/vnd.apache.parquet", **hub_response_body["headers"]},
                content=buffer.getvalue(),
                auth=None,
                timeout=timeout,  # required for large size dataset
            )
            bucket_response.raise_for_status()

        else:
            hub_response.raise_for_status()

        logger.success(
            "Your dataset has been successfully uploaded to the Hub. "
            f"View it here: {urljoin(self.settings.hub_url, f'datasets/{dataset.owner}/{dataset.name}')}"
        )

        return response

    def upload_benchmark(self, benchmark: BenchmarkSpecification, access: AccessType = "private"):
        """Upload the benchmark to the Polaris Hub.

        Info: Owner
            You have to manually specify the owner in the benchmark data model. Because the owner could
            be a user or an organization, we cannot automatically infer this from the logged-in user.

        Note: Required meta-data
            The Polaris client and hub maintain different requirements as to which meta-data is required.
            The requirements by the hub are stricter, so when uploading to the hub you might
            get some errors on missing meta-data. Make sure to fill-in as much of the meta-data as possible
            before uploading.

        Note: Non-existent datasets
            The client will _not_ upload the associated dataset to the hub if it does not yet exist.
            Make sure to specify an existing dataset or upload the dataset first.

        Args:
            benchmark: The benchmark to upload.
            access: Grant public or private access to result
        """

        # Get the serialized data-model
        # We exclude the dataset as we expect it to exist on the hub already.
        benchmark_json = benchmark.model_dump(exclude={"dataset"}, exclude_none=True, by_alias=True)
        benchmark_json["datasetArtifactId"] = benchmark.dataset.artifact_id
        benchmark_json["access"] = access

        url = f"/benchmark/{benchmark.owner}/{benchmark.name}"
        response = self._base_request_to_hub(url=url, method="PUT", json=benchmark_json)

        logger.success(
            "Your benchmark has been successfully uploaded to the Hub. "
            f"View it here: {urljoin(self.settings.hub_url, f'benchmarks/{benchmark.owner}/{benchmark.name}')}"
        )
        return response



class InvalidDatasetError(ValueError):
    pass


class InvalidBenchmarkError(ValueError):
    pass


class InvalidResultError(ValueError):
    pass


class PolarisChecksumError(ValueError):
    pass


class PolarisHubError(Exception):
    pass


class PolarisUnauthorizedError(PolarisHubError):
    DEFAULT_ERROR_MSG = (
        "You are not logged in to Polaris or your login has expired. "
        "You can use the Polaris CLI to easily authenticate yourself again, see `polaris login --help`."
    )

    def __init__(self, message: str = DEFAULT_ERROR_MSG):
        super().__init__(message)


class TestAccessError(Exception):
    # Prevent pytest to collect this as a test
    __test__ = False

    pass



from typing import List
import seaborn as sns
import matplotlib.pyplot as plt
from loguru import logger
from scipy import stats
import datamol as dm

from .utils import PandasDataFrame
from ._chemistry_curator import NUM_UNDEF_STEREO_CENTER, NUM_DEF_STEREO_CENTER


def visualize_distribution(data_cols: List[str], dataset: PandasDataFrame):
    """Visualize the distribution of the data and highlight the potential outliers.

    Args:
        data_cols: Column names which the `OUTLIER` was previously detected.
        dataset: Dataset Dataframe which inludes columns of `OUTLIER` flags.

    Returns:
        figs: List of distribution plots.
    """
    figs = []
    for col in data_cols:
        outlier_col = f"OUTLIER_{col}"
        if outlier_col not in dataset.columns:
            raise ValueError(
                f"{outlier_col} is not found in the given dataset.\
                             Please run <polaris.curation.utils.outlier_detection> first."
            )
        fig, axes = plt.subplots(ncols=2, nrows=1, figsize=(16, 4))
        to_plot = dataset.sort_values(by=col).reset_index()
        to_plot["order"] = range(to_plot.shape[0])
        sns.scatterplot(
            to_plot,
            x="order",
            y=col,
            hue=outlier_col,
            palette={True: "red", False: "navy"},
            ax=axes[0],
        )
        stats.probplot(to_plot[col].values, dist="norm", plot=axes[1])
        figs.append(fig)
    return figs


def verify_stereoisomers(data_cols: List[str], dataset: PandasDataFrame):
    """Visual verification the stereoisomers from the original dataset

    Args:
        data_cols: Column names which the activity cliff was previously detected.
        dataset: Dataframe which includes stereo chemistry information of the molcules.

    """
    figs = []
    for col in data_cols:
        logger.info(f"Verify the stereo ismomers for readout `{col}`")
        cliff_col = f"CLASS_{col}_stereo_cliff"
        if cliff_col not in dataset:
            raise ValueError(
                "The stereo chemistry information is unavailable. \
                             Please run the <polaris.curation.run_chemistry_curation> first."
            )
        cliff = dataset[cliff_col].notna()
        if cliff.sum() > 0:
            to_plot = dataset.loc[
                cliff, ["smiles", NUM_UNDEF_STEREO_CENTER, NUM_DEF_STEREO_CENTER, col, cliff_col]
            ]
            fig = dm.to_image([dm.to_mol(s) for s in to_plot.smiles])
            figs.append(fig)
        else:
            logger.info("No activity cliffs found in stereosimoers.")


def check_undefined_stereocenters(dataset: PandasDataFrame):
    """Visual check the molecules with undefined stereo centers

    Args:
        dataset: Dataframe which includes stereo chemistry information of the molcules.

    Returns:
        fig: Image figure of the molecules with stereo center annotations.

    See Also:
        <polaris.curation.run_chemistry_curation>

    """

    if NUM_UNDEF_STEREO_CENTER not in dataset or NUM_DEF_STEREO_CENTER not in dataset:
        raise ValueError(
            "The stereo information of the molecules are unavailable.\
                         Please run the <polaris.curation.run_chemistry_curation> first."
        )

    to_plot = dataset[dataset[[NUM_UNDEF_STEREO_CENTER, NUM_DEF_STEREO_CENTER]].notna().any(axis=1)]
    legends = (
        to_plot[[NUM_UNDEF_STEREO_CENTER, NUM_DEF_STEREO_CENTER]]
        .apply(
            lambda x: f"num_undefined_stereo_center:{x[NUM_UNDEF_STEREO_CENTER]}\n num_defined_stereo_center{x[NUM_DEF_STEREO_CENTER]}",
            axis=1,
        )
        .tolist()
    )
    fig = dm.to_image(to_plot.mol.tolist(), legends=legends)
    return fig



# class to perform data curation for both chemistry and endpoint measured values
from typing import List, Optional

from pydantic import BaseModel

from ._chemistry_curator import run_chemistry_curation
from ._data_curator import run_data_curation
from .utils import PandasDataFrame

ORI_PREFIX = "ORIGINAL_"


class MolecularCurator(BaseModel):
    """
    Apply curation process on given dataset on both chemistry and endpoint measurements

    Args:
        data: Data frame which contains molecule and bioactivity data columns
        mol_col: Molecule column name
        data_cols: Bioactivity data column names
        mask_stereo_undefined_mols: If set to 'True', for the molecule which contains undefined stereochemistry
            center, the bioactivity data will be set to 'Nan' and be ignored.
        ignore_stereo: If set to 'True', the stereochemistry information will be ignored from data processing steps,
            such as detect replicated molecules, detection of activity cliff.
        class_thresholds: Dictionary of bioactivity column names and the thresholds for discretizing
            the continuous data.
        outlier_params: Parameters for automatic outlier detection.
        progress: Whether show progress of the parallelization process.

    See Also:
        <polaris.curation._data_curator.run_data_curation>
        <polaris.curation._chemistry_curator.run_chemistry_curation>
    """

    data: PandasDataFrame
    mol_col: str = "smiles"
    data_cols: List[str]
    mask_stereo_undefined_mols: bool = False
    ignore_stereo: bool = False
    class_thresholds: Optional[dict] = None
    outlier_params: Optional[dict] = None
    activity_cliff_params: Optional[dict] = None
    progress: bool = False

    def __call__(self):
        # copy the original data columns
        data = self.data.copy()
        for data_col in self.data_cols + [self.mol_col]:
            data[f"{ORI_PREFIX}{data_col}"] = self.data[data_col].copy()

        # run chemistry curation
        dataframe = run_chemistry_curation(
            self.data[self.mol_col], ignore_stereo=self.ignore_stereo, progress=False
        )

        for col in dataframe.columns:
            data[col] = dataframe[col].values

        # run endpoint curation
        data = run_data_curation(
            data=data,
            data_cols=self.data_cols,
            mask_stereo_undefined_mols=self.mask_stereo_undefined_mols,
            ignore_stereo=self.ignore_stereo,
            class_thresholds=self.class_thresholds,
            outlier_params=self.outlier_params,
            activity_cliff_params=self.activity_cliff_params,
        )

        data = data.reset_index(drop=True)
        return data



from typing import Any

from polaris.utils.types import SlugCompatibleStringType


def listit(t: Any):
    """
    Converts all tuples in a possibly nested object to lists
    https://stackoverflow.com/questions/1014352/how-do-i-convert-a-nested-tuple-of-tuples-and-lists-to-lists-of-lists-in-python
    """
    return list(map(listit, t)) if isinstance(t, (list, tuple)) else t


def sluggify(sluggable: SlugCompatibleStringType):
    """
    Converts a string to a slug-compatible string.
    """
    return sluggable.lower().replace("_", "-")



from typing import Dict, List

import pandas as pd
from loguru import logger
from sklearn.utils.multiclass import type_of_target

from ._chemistry_curator import NO_STEREO_UNIQUE_ID, UNDEF_EZ, UNDEF_ED, UNIQUE_ID
from .utils import discretizer, modified_zscore, outlier_detection

CATEGORIES = ["binary", "multiclass"]
CONTINUOUS = ["continuous"]
CLASS_PREFIX = "CLASS_"


def _identify_stereoisomers_with_activity_cliff(
    data: pd.DataFrame, data_col: str, threshold: float = 1, groupby_col: str = NO_STEREO_UNIQUE_ID
) -> List:
    """Identify the stereoisomers which show activity cliff.
       For continuous data, the activity cliff is defined by the `zscore` difference greater than 1.
       For categorical data, the activity cliff is defined by class labels for categorical values.
       It's recommended to provide clear thresholds to classify the bioactivity if possible.

    Args:
        data: Dataframe which contains unique ID which identifies the stereoisomers.
        data_col: Column name for the targeted activity
        threshold: Zscore threshold which defines the activity cliff.
                   By default, for continuous values, the activity cliff is defined as if the difference between zscores
                   is greater than 1.
        groupby_col: Column name which can be used to identifier the stereoisomers.

    Returns:
        mol_hash_ids_with_cliff: A list mol hash ids for the stereoisomer molecules which show activity cliff.

    """
    data_type = type_of_target(data[data_col].dropna().values, input_name=data_col)

    if data_type is None:
        raise ValueError(f"The column {data_col} contains less than 2 unique values.")

    mol_hash_ids_with_cliff = []
    if data_type in CONTINUOUS:
        # compute modified zscore
        data[f"{data_col}_zscore"] = modified_zscore(data[data_col].values)[0]

    for hashmol_id, group_df in data.groupby(groupby_col):
        if group_df.shape[0] > 1:
            if data_type in CONTINUOUS:
                has_cliff = (
                    group_df[f"{data_col}_zscore"].max() - group_df[f"{data_col}_zscore"].min()
                ) > threshold
            else:
                has_cliff = len(group_df[data_col].unique()) > 1

            if has_cliff:
                mol_hash_ids_with_cliff.append(hashmol_id)

    return mol_hash_ids_with_cliff


def _process_stereoisomer_with_activity_cliff(
    data: pd.DataFrame, data_cols: List[str], mask_stereo_undefined_mols: bool = True, **kwargs
):
    """Identify and discard the stereoisomers which the stereocenter are not well-defined but show activity cliff.

    Args:
        data: Parsed dataset
        data_cols: Data column names to be processed.
        mask_stereo_undefined_mols: Whether remove molecules and the bioactivity which shows chiral shift
                                    but stereo information not defined in the molecule representation.
        kwargs: Parameters for identify the activity cliff between the stereoisomers.
    """
    for data_col in data_cols:
        data_col_mask = (
            [data_col, data_col[len(CLASS_PREFIX) :]] if data_col.startswith(CLASS_PREFIX) else data_col
        )
        mol_with_cliff = _identify_stereoisomers_with_activity_cliff(data=data, data_col=data_col, **kwargs)
        data.loc[data[NO_STEREO_UNIQUE_ID].isin(mol_with_cliff), f"{data_col}_stereo_cliff"] = True
        if len(mol_with_cliff) > 0 and mask_stereo_undefined_mols:
            mol_ids = data.query(f"`{data_col}_stereo_cliff`==True & (`{UNDEF_ED}` | `{UNDEF_EZ}`) ")[
                UNIQUE_ID
            ].values
            data.loc[data[UNIQUE_ID].isin(mol_ids), data_col_mask] = None
    return data


def _merge_duplicates(
    data: pd.DataFrame, data_cols: List[str], merge_on=None, keep_all_rows: bool = False
) -> pd.DataFrame:
    """Merge molecules with multiple measurements.
       To be robust to the outliers, `median` is used to compute the average value cross all measurements.

    Args:
        data: Parsed dataset
        data_cols: Data column names to be processed
        merge_on: Column name to identify the duplicated molecules.
                  dy default, `molhash_id` is used to identify the replicated molecule.

    """
    if merge_on is None:
        merge_on = [UNIQUE_ID]

    df_list = []
    # Include 'molhash without stereo information' additionally for easily tracking the stereoisomers
    for uid, df in data.groupby(by=merge_on):
        data_vals = df[data_cols].median(axis=0, skipna=True).tolist()
        df.loc[:, data_cols] = data_vals
        df_list.append(df)
    merged_df = pd.concat(df_list).sort_values(by=merge_on)
    if not keep_all_rows:
        merged_df = merged_df.drop_duplicates(subset=merge_on, keep="first").reset_index(drop=True)
    return merged_df


def _class_conversion(
    data: pd.DataFrame,
    data_cols: List[str],
    conversion_params: List[dict],
    prefix: str = CLASS_PREFIX,
) -> pd.DataFrame:
    """
    Apply binary or multiclass conversion to the data columns by the given thresholds

    Args:
        data: Parsed dataset
        data_cols: Data column names to be converted
        conversion_params: List of parameters for class conversion
            '''
            Example:
                [
                 # for data column 1 listed in `data_cols` - binary class
                 {"threshold_values": [0.5], "threshold_labels": [0, 1],"nan_value": "NA" },
                 # for data column 2 listed in `data_cols` - multiclass
                 {"threshold_values": [10, 20], "threshold_labels": [0, 1, 2],"nan_value": "-"}
                ]

            '''
        prefix: Prefix for the added column names

    Returns:
        dataset: Dataset with clss columns
    """
    for data_col in data_cols:
        thresholds = conversion_params.get(data_col, None)
        if thresholds is not None:
            data[f"{prefix}{data_col}"] = discretizer(X=data[data_col].values, allow_nan=True, **thresholds)
    return data


def check_outliers(data, data_cols, method: str = "zscore", prefix="OUTLIER", **kwargs) -> pd.DataFrame:
    """Automatic detection of outliers of bioactivity

    Args:
        data: Parsed dataset
        data_cols: Data column names for outlier detections
        method: Algorithm name of outlier detection. The supported methods are
            "iso": IsolationForest
            "lof": LocalOutlierFactor
            "svm": OneClassSVM
            "ee": EllipticEnvelope
            "zscore": ZscoreOutlier
        kwargs: Additional parameters for the automatic outlier detection method.
                The parameters impact largely the detection therefore should be carefully chosen.
        prefix: Prefix for the boolean column which indicate whether the data point is an outlier.

    Returns:
        Dataframe with added boolean column for outliers.

    See also:
        <sklearn.ensemble.IsolationForest>
        <sklearn.svm.OneClassSVM>
        <sklearn.covariance.EllipticEnvelope>
        <sklearn.neighbors.LocalOutlierFactor>
        <polaris.curation.util.ZscoreOutlier>

    """
    for data_col in data_cols:
        outlier_col = f"{prefix}_{data_col}"
        data[f"{prefix}_{data_col}"] = None
        ind = data.index.values
        notna_ind = ind[data[data_col].notna()]
        data.loc[notna_ind, outlier_col] = False
        outlier_ind = outlier_detection(X=data[data_col].dropna().values, method=method, **kwargs)[:, 0]
        outlier_ind = notna_ind[outlier_ind]

        data.loc[outlier_ind, outlier_col] = True

        if len(outlier_ind) > 0:
            logger.warning(
                f"Detected {len(outlier_ind)} outliers for data column {data_col} using {method}. "
                f"Please revise the data and consider remove the outliers. "
            )
    return data


def run_data_curation(
    data: pd.DataFrame,
    data_cols: List[str],
    mask_stereo_undefined_mols: bool = False,
    ignore_stereo: bool = False,
    class_thresholds: Dict = None,
    outlier_params: Dict = None,
    activity_cliff_params: Dict = None,
) -> pd.DataFrame:
    """Perform curation on the measured values from biological assay

    Args:
        data: Dataframe which include molecule, measured values.
        data_cols: Column names for the data of interest.
        mask_stereo_undefined_mols: Whether mask the molecule stereo pairs which show activity cliff from dataset.
        ignore_stereo: Ignore the stereochemistry information from data curation.
        class_thresholds: Parameters to define class labels for the above listed `data_cols`.
        outlier_params: Parameters for outlier detection.
        activity_cliff_params: Parameter for identify the activity cliff between the stereoisomers.

    See Also:

    """
    data = check_outliers(data, data_cols, **outlier_params if outlier_params is not None else {})

    # User should deal with scaling and normalization on their end
    data = _merge_duplicates(
        data=data,
        data_cols=data_cols,
        merge_on=[NO_STEREO_UNIQUE_ID] if ignore_stereo else [UNIQUE_ID],
    )
    # class conversion
    if class_thresholds:
        data = _class_conversion(data, data_cols, class_thresholds, prefix=CLASS_PREFIX)
    if not ignore_stereo:
        # detect stereo activity cliff, keep or remove
        data = _process_stereoisomer_with_activity_cliff(
            data=data,
            data_cols=[
                f"{CLASS_PREFIX}{data_col}" if f"{CLASS_PREFIX}{data_col}" in data.columns else data_col
                for data_col in data_cols
            ]
            if class_thresholds
            else data_cols,
            mask_stereo_undefined_mols=mask_stereo_undefined_mols,
            **activity_cliff_params if activity_cliff_params is not None else {},
        )

    return data



import os

import numpy as np
import pandas as pd

from polaris.benchmark import MultiTaskBenchmarkSpecification, SingleTaskBenchmarkSpecification
from polaris.evaluate._metric import Metric
from polaris.evaluate._results import BenchmarkResults
from polaris.utils.types import HubOwner


def test_result_to_json(tmpdir: str, test_user_owner: HubOwner):
    scores = pd.DataFrame(
        {
            "Test set": ["A", "A", "A", "A", "B", "B", "B", "B"],
            "Target label": ["C", "C", "D", "D", "C", "C", "D", "D"],
            "Metric": ["mean_absolute_error", "accuracy"] * 4,
            "Score": [0.1] * 8,
        }
    )

    result = BenchmarkResults(
        name="test",
        description="Lorem ipsum!",
        tags=["test"],
        user_attributes={"key": "value"},
        owner=test_user_owner,
        results=scores,
        benchmark_name="my-benchmark",
        benchmark_owner=test_user_owner,
        github_url="https://github.com/",
        paper_url="https://chemrxiv.org/",
        contributors=["my-user", "other-user"],
    )

    path = os.path.join(tmpdir, "result.json")
    result.to_json(path)
    BenchmarkResults.from_json(path)


def test_metrics_singletask_reg(tmpdir: str, test_single_task_benchmark: SingleTaskBenchmarkSpecification):
    _, test = test_single_task_benchmark.get_train_test_split()
    predictions = np.random.random(size=test.inputs.shape[0])
    result = test_single_task_benchmark.evaluate(predictions)
    assert isinstance(result.results, pd.DataFrame)
    assert set(result.results.columns) == {"Test set", "Target label", "Metric", "Score"}
    for metric in test_single_task_benchmark.metrics:
        assert metric in result.results.Metric.tolist()


def test_metrics_multitask_reg(tmpdir: str, test_multi_task_benchmark: MultiTaskBenchmarkSpecification):
    train, test = test_multi_task_benchmark.get_train_test_split()
    predictions = {
        target_col: np.random.random(size=test.inputs.shape[0]) for target_col in train.target_cols
    }
    result = test_multi_task_benchmark.evaluate(predictions)
    for metric in test_multi_task_benchmark.metrics:
        assert metric in result.results.Metric.tolist()


def test_metrics_singletask_clf(
    tmpdir: str, test_single_task_benchmark_clf: SingleTaskBenchmarkSpecification
):
    _, test = test_single_task_benchmark_clf.get_train_test_split()
    predictions = np.random.randint(2, size=test.inputs.shape[0])
    result = test_single_task_benchmark_clf.evaluate(predictions)
    for metric in test_single_task_benchmark_clf.metrics:
        assert metric in result.results.Metric.tolist()


def test_metrics_multitask_clf(tmpdir: str, test_multi_task_benchmark_clf: MultiTaskBenchmarkSpecification):
    train, test = test_multi_task_benchmark_clf.get_train_test_split()
    predictions = {
        target_col: np.random.randint(2, size=test.inputs.shape[0]) for target_col in train.target_cols
    }
    result = test_multi_task_benchmark_clf.evaluate(predictions)
    assert isinstance(result.results, pd.DataFrame)
    assert set(result.results.columns) == {"Test set", "Target label", "Metric", "Score"}
    # check the targets and metrics
    assert result.results.shape[0] == len(test_multi_task_benchmark_clf.target_cols) * len(
        test_multi_task_benchmark_clf.metrics
    )
    for metric in test_multi_task_benchmark_clf.metrics:
        assert metric in result.results.Metric.tolist()


def test_metric_direction():
    for metric in Metric:
        assert metric.value.direction in ["min", "max"]



import datamol as dm
import numpy as np
import pandas as pd
import pydantic
import pytest

from polaris.curation import run_chemistry_curation
from polaris.curation._chemistry_curator import _num_stereo_centers
from polaris.curation._data_curator import (
    _class_conversion,
    _identify_stereoisomers_with_activity_cliff,
    _merge_duplicates,
    check_outliers,
)
from polaris.curation.utils import Discretizer, LabelOrder, discretizer, outlier_detection


def test_discretizer():
    X = [[1.0, -1.0, 2.0], [2.0, 0.0, 0.0], [0.0, 1.0, -1.0]]
    thresholds_binary = [0.5]
    thresholds_multiclass = [0, 1]

    values_binary = discretizer(X=X, thresholds=thresholds_binary)
    assert np.array_equal(values_binary, np.array([[1, 0, 1], [1, 0, 0], [0, 1, 0]]))

    values_binary_r = discretizer(X=X, thresholds=thresholds_binary, label_order=LabelOrder.desc.value)
    assert np.array_equal(values_binary_r, np.array([[0, 1, 0], [0, 1, 1], [1, 0, 1]]))

    values_multiclass = discretizer(X=X, thresholds=thresholds_multiclass)
    assert np.array_equal(values_multiclass, np.array([[2, 0, 2], [2, 1, 1], [1, 2, 0]]))

    values_multiclass_r = discretizer(
        X=X, thresholds=thresholds_multiclass, label_order=LabelOrder.desc.value
    )
    assert np.array_equal(values_multiclass_r, np.array([[0, 2, 0], [0, 1, 1], [1, 0, 2]]))

    with pytest.raises(pydantic.ValidationError):
        Discretizer(thresholds=thresholds_multiclass, label_order="WrongType")

    with pytest.raises(ValueError):
        discretizer(X=X, thresholds=thresholds_multiclass, label_order="WrongType")


def test_class_conversion():
    colnames = ["data_col_1", "data_col_2", "data_col_3"]
    data = pd.DataFrame.from_dict(
        {colnames[0]: [1.0, -1.0, 2.0], colnames[1]: [2.0, 0.0, 0.0], colnames[2]: [0.0, 1.0, -1.0]}
    )
    converted = _class_conversion(
        data,
        data_cols=colnames,
        conversion_params={
            "data_col_1": {"thresholds": [0, 1], "label_order": "ascending"},
            "data_col_2": {"thresholds": [0.5], "label_order": LabelOrder.acs.value},
            "data_col_3": {"thresholds": [0, 1], "label_order": LabelOrder.desc.value},
        },
    )
    assert converted["CLASS_data_col_1"].tolist() == [2, 0, 2]
    assert converted["CLASS_data_col_2"].tolist() == [1, 0, 0]
    assert converted["CLASS_data_col_3"].tolist() == [1, 0, 2]


def test_run_chemistry_curation():
    mols = [
        "COc1ccc2ncc(=O)n(CCN3CC[C@H](NCc4ccc5c(n4)NC(=O)CO5)[C@H](O)C3)c2c1",
        "COc1ccc2ncc(=O)n(CCN3CC[C@@H](NCc4ccc5c(n4)NC(=O)CO5)[C@@H](O)C3)c2c1",
        "C[C@H]1CN(Cc2cc(Cl)ccc2OCC(=O)O)CCN1S(=O)(=O)c1ccccc1",
        "C[C@@H]1CN(Cc2cc(Cl)ccc2OCC(=O)O)CCN1S(=O)(=O)c1ccccc1",
        "CC[C@@H](c1ccc(C(=O)O)c(Oc2cccc(Cl)c2)c1)N1CCC[C@H](n2cc(C)c(=O)[nH]c2=O)C1",
        "CC[C@H](c1ccc(C(=O)O)c(Oc2cccc(Cl)c2)c1)N1CCC[C@H](n2cc(C)c(=O)[nH]c2=O)C1",
        "CNc1cc(Nc2cccn(-c3ccccn3)c2=O)nn2c(C(=O)N[C@@H]3C[C@@H]3F)cnc12",
        "CCOc1cc2nn(CCC(C)(C)O)cc2cc1NC(=O)c1cccc(C(F)F)n1",
        "CN(c1ncc(F)cn1)C1CCCNC1",
        "CC(C)(Oc1ccc(-c2cnc(N)c(-c3ccc(Cl)cc3)c2)cc1)C(=O)O",
        "CC(C)(O)CCn1cc2cc(NC(=O)c3cccc(C(F)(F)F)n3)c(C(C)(C)O)cc2n1",
        "CC#CC(=O)N[C@H]1CCCN(c2c(F)cc(C(N)=O)c3[nH]c(C)c(C)c23)C1",
        "COc1ccc(Cl)cc1C(=O)NCCc1ccc(S(=O)(=O)NC(=O)NC2CCCCC2)cc1",
        "C=CC(=O)N1CCC[C@@H](n2nc(-c3ccc(Oc4ccccc4)cc3)c3c(N)ncnc32)C1",
        "CC(C)NC(=O)COc1cccc(-c2nc(Nc3ccc4[nH]ncc4c3)c3ccccc3n2)c1.[Na]",
        "CCS(=O)(=O)N1CC(CC#N)(n2cc(-c3ncnc4[nH]ccc34)cn2)C1.Cl",
    ]
    # check stereoisomers are included.
    df = run_chemistry_curation(mols=mols)
    assert len(df.smiles.unique()) == len(mols)

    # check if stereoisomers are ignored
    df = run_chemistry_curation(mols=mols, ignore_stereo=True)
    assert len(df.smiles.unique()) == len(mols) - 3

    # check whether salts/solvents were removed.
    for smiles in df.smiles:
        mol = dm.to_mol(smiles)
        assert dm.same_mol(dm.remove_salts_solvents(mol), mol)


def test_check_outlier_zscore():
    data = pd.DataFrame(np.random.normal(0, 0.1, size=(100, 1)), columns=["data_col"])
    # create 5 data points which are out of distribution
    num_outlier = 5
    data.loc[: num_outlier - 1, "data_col"] = 10
    outilers = outlier_detection(X=data[["data_col"]].values, method="zscore")
    assert len(outilers) == num_outlier

    data_outlier = check_outliers(data, ["data_col"])
    assert data_outlier["OUTLIER_data_col"].sum() == num_outlier


def test_identify_stereoisomers_with_activity_cliff():
    vals = list(np.random.randint(0, 10, 50)) * 2 + np.random.normal(0, 0.01, 100)
    groups = list(range(50)) * 2
    num_cliff = 10
    index_cliff = np.random.randint(0, 50, num_cliff)
    vals[index_cliff] = 1000

    data = pd.DataFrame({"data_col": vals, "groupby_col": groups})
    ids = _identify_stereoisomers_with_activity_cliff(
        data=data, data_col="data_col", groupby_col="groupby_col", threshold=1
    )
    # check if identifed ids are correct
    assert set(ids) == set(index_cliff)


def test_merge_duplicates():
    num_sample = 100
    ids = np.array(range(num_sample))
    max_ind = max(ids)
    num_dup = 5
    for index in range(num_dup):
        ids[max_ind - index] = index
    data_col_1 = np.random.normal(1, 0.01, 100)
    data_col_2 = np.random.normal(2, 0.01, 100)
    data_col_3 = np.random.normal(3, 0.01, 100)
    data = pd.DataFrame(
        {"data_col_1": data_col_1, "data_col_2": data_col_2, "data_col_3": data_col_3, "ids": ids}
    )

    merged = _merge_duplicates(
        data=data, data_cols=["data_col_1", "data_col_2", "data_col_3"], merge_on=["ids"]
    )

    # check the data points been merged
    assert data.shape[0] == merged.shape[0] + num_dup
    # check the merged values are correct
    for index in range(num_dup):
        assert data.loc[[index, max_ind - index], "data_col_1"].median() == merged.loc[index, "data_col_1"]
        assert data.loc[[index, max_ind - index], "data_col_2"].median() == merged.loc[index, "data_col_2"]
        assert data.loc[[index, max_ind - index], "data_col_3"].median() == merged.loc[index, "data_col_3"]


def test_num_undefined_stereo_centers():
    # mol with no stereo centers
    mol = dm.to_mol("CCCC")
    num_all, num_defined, num_undefined = _num_stereo_centers(mol)
    assert num_all == 0
    assert num_defined == 0
    assert num_undefined == 0

    # mol with all defined centers
    mol = dm.to_mol("C1C[C@H](C)[C@H](C)[C@H](C)C1")
    num_all, num_defined, num_undefined = _num_stereo_centers(mol)
    assert num_all == 3
    assert num_defined == 3
    assert num_undefined == 0

    # mol with partial defined centers
    mol = dm.to_mol("C[C@H](F)C(F)(Cl)Br")
    num_all, num_defined, num_undefined = _num_stereo_centers(mol)
    assert num_all == 2
    assert num_defined == 1
    assert num_undefined == 1

    # mol with no defined centers
    mol = dm.to_mol("CC(F)C(F)(Cl)Br")
    num_all, num_defined, num_undefined = _num_stereo_centers(mol)
    assert num_all == 2
    assert num_defined == 0
    assert num_undefined == 2



import fsspec
import yaml

from polaris.benchmark._definitions import (
    MultiTaskBenchmarkSpecification,
    SingleTaskBenchmarkSpecification,
)
from polaris.dataset._dataset import Dataset
from polaris.hub.client import PolarisHubClient
from polaris.utils import fs
from polaris.utils.errors import InvalidBenchmarkError, InvalidDatasetError


def load_dataset(path: str):
    """
    Loads the dataset. Inspired by the HF API, this can either load from a remote or local path or from the Hub.
    """

    extension = fs.get_extension(path)
    is_file = fs.is_file(path) or extension == "zarr"

    if not is_file:
        # Load from the Hub
        client = PolarisHubClient()
        return client.get_dataset(*path.split("/"))

    if extension == "zarr":
        return Dataset.from_zarr(path)
    elif extension == "json":
        return Dataset.from_json(path)

    raise NotImplementedError("This should not be reached.")


def load_benchmark(path: str):
    """
    Loads a benchmark.
    """

    is_file = fs.is_file(path) or fs.get_extension(path) == "zarr"

    if not is_file:
        # Load from the Hub
        client = PolarisHubClient()
        return client.get_benchmark(*path.split("/"))

    with fsspec.open(path, "r") as fd:
        data = yaml.safe_load(fd)  # type: ignore

    # TODO (cwognum): As this gets more complex, how do we effectivly choose which class we should use?
    #  e.g. we might end up with a single class per benchmark.
    is_single_task = isinstance(data["target_cols"], str) or len(data["target_cols"]) == 1
    cls = SingleTaskBenchmarkSpecification if is_single_task else MultiTaskBenchmarkSpecification
    return cls.from_json(path)



from typing import Optional

import typer

from polaris.hub.client import PolarisHubClient

app = typer.Typer(
    add_completion=False,
    help="Polaris is a framework for benchmarking methods in drug discovery.",
)


@app.command("login", help="Log in to the Polaris Hub")
def login(
    client_env_file: Optional[str] = None,
    auto_open_browser: bool = True,
    overwrite: bool = False,
):
    """Uses the OAuth2 protocol to gain token-based access to the Polaris Hub API"""
    with PolarisHubClient(env_file=client_env_file) as client:
        client.login(auto_open_browser=auto_open_browser, overwrite=overwrite)


@app.command(hidden=True)
def secret():
    # NOTE (cwognum): Empty, hidden command to force Typer to not collapse the subcommand.
    # Added because I anticipate we will want to add more subcommands later on. This will keep
    # the API consistent in the meantime. Once there are other subcommands, it can be removed.
    # See also: https://github.com/tiangolo/typer/issues/315
    raise NotImplementedError()


if __name__ == "__main__":
    app()



from polaris.dataset._column import ColumnAnnotation
from polaris.dataset._dataset import Dataset
from polaris.dataset._subset import Subset

__all__ = [
    "ColumnAnnotation",
    "Dataset",
    "Subset",
]



from pydantic import field_validator

from polaris.benchmark._base import BenchmarkSpecification


class SingleTaskBenchmarkSpecification(BenchmarkSpecification):
    """Subclass for any single-task benchmark specification

    In addition to the data-model and logic of the base-class,
    this class verifies that there is just a single target-column.
    """

    @field_validator("target_cols", check_fields=False)
    def validate_target_cols(cls, v):
        if not len(v) == 1:
            raise ValueError("A single-task benchmark should specify a single target column")
        return v


class MultiTaskBenchmarkSpecification(BenchmarkSpecification):
    """Subclass for any multi-task benchmark specification

    In addition to the data-model and logic of the base-class,
    this class verifies that there are multiple target-columns.
    """

    @field_validator("target_cols", check_fields=False)
    def validate_target_cols(cls, v):
        if not len(v) > 1:
            raise ValueError("A multi-task benchmark should specify at least two target columns")
        return v



"""
The `fs` module makes it easier to work with all type of path (the ones supported by `fsspec`).
"""

import hashlib
import io
import os
import pathlib
from typing import List, Optional, Union

import fsspec
import fsspec.utils
from datamol.utils import parallelized


def _import_tqdm():
    try:
        from tqdm.auto import tqdm

        return tqdm
    except ImportError:
        return None


def get_mapper(path: Union[str, os.PathLike]):
    """Get the fsspec mapper.

    Args:
        path: a path supported by `fsspec` such as local, s3, gcs, etc.
    """
    return fsspec.get_mapper(str(path))


def get_basename(path: Union[str, os.PathLike]):
    """Get the basename of a file or a folder.

    Args:
        path: a path supported by `fsspec` such as local, s3, gcs, etc.
    """
    path = str(path)
    mapper = get_mapper(path)
    clean_path = path.rstrip(mapper.fs.sep)
    return str(clean_path).split(mapper.fs.sep)[-1]


def get_extension(path: Union[str, os.PathLike]):
    """Get the extension of a file.

    Args:
        path: a path supported by `fsspec` such as local, s3, gcs, etc.
    """
    basename = get_basename(path)
    return basename.split(".")[-1]


def exists(path: Union[str, os.PathLike, fsspec.core.OpenFile, io.IOBase]):
    """Check whether a file or a directory exists.

    Important: File-like object always exists.

    Args:
        path: a path supported by `fsspec` such as local, s3, gcs, etc.
    """
    return is_file(path) or is_dir(path)


def is_file(path: Union[str, os.PathLike, fsspec.core.OpenFile, io.IOBase]):
    """Check whether a file exists.

    Args:
        path: a path supported by `fsspec` such as local, s3, gcs, etc.
    """
    if isinstance(path, fsspec.core.OpenFile):
        return path.fs.isfile(path.path)

    elif isinstance(path, (str, os.PathLike)):
        mapper = get_mapper(str(path))
        return mapper.fs.isfile(str(path))

    else:
        return False


def is_dir(path: Union[str, os.PathLike, fsspec.core.OpenFile, io.IOBase]):
    """Check whether a file exists.

    Args:
        path: a path supported by `fsspec` such as local, s3, gcs, etc.
    """
    if isinstance(path, fsspec.core.OpenFile):
        return path.fs.isdir(path.path)

    elif isinstance(path, (str, os.PathLike)):
        mapper = get_mapper(str(path))
        return mapper.fs.isdir(str(path))

    else:
        return False


def get_protocol(path: Union[str, os.PathLike], fs: Optional[fsspec.AbstractFileSystem] = None):
    """Return the name of the path protocol.

    Args:
        path: a path supported by `fsspec` such as local, s3, gcs, etc.
    """

    if fs is None:
        fs = get_mapper(path).fs

    protocol = fs.protocol  # type: ignore

    if "s3" in protocol:
        return "s3"
    elif "gs" in protocol:
        return "gs"
    elif isinstance(protocol, (tuple, list)):
        return protocol[0]
    return protocol


def is_local_path(path: Union[str, os.PathLike]):
    """Check whether a path is local."""
    return get_protocol(str(path)) == "file"


def join(*paths: str):
    """Join paths together. The first element determine the
    filesystem to use (and so the separator.

    Args:
        *paths: a list of paths supported by `fsspec` such as local, s3, gcs, etc.
    """
    _paths = [str(path).rstrip("/") for path in paths]
    source_path = _paths[0]
    fs = get_mapper(source_path).fs
    full_path = fs.sep.join(_paths)
    return full_path


def get_size(file: Union[str, os.PathLike, io.IOBase, fsspec.core.OpenFile]) -> Optional[int]:
    """Get the size of a file given its path. Return None if the
    size can't be retrieved.
    """

    if isinstance(file, io.IOBase) and hasattr(file, "name"):
        fs_local = fsspec.filesystem("file")
        file_size = fs_local.size(getattr(file, "name"))

    elif isinstance(file, (str, os.PathLike)):
        fs = get_mapper(str(file)).fs
        file_size = fs.size(str(file))

    elif isinstance(file, fsspec.core.OpenFile):
        file_size = file.fs.size(file.path)

    else:
        file_size = None

    return file_size


def copy_file(
    source: Union[str, pathlib.Path, io.IOBase, fsspec.core.OpenFile],
    destination: Union[str, pathlib.Path, io.IOBase, fsspec.core.OpenFile],
    chunk_size: Optional[int] = None,
    force: bool = False,
    progress: bool = False,
    leave_progress: bool = True,
):
    """Copy one file to another location across different filesystem (local, S3, GCS, etc).

    Args:
        source: path or file-like object to copy from.
        destination: path or file-like object to copy to.
        chunk_size: the chunk size to use. If progress is enabled the chunk
            size is `None`, it is set to 1MB (1024 * 1024).
        force: whether to overwrite the destination file if it exists.
        progress: whether to display a progress bar.
        leave_progress: whether to hide the progress bar once the copy is done.
    """

    if progress and chunk_size is None:
        chunk_size = 1024 * 1024

    if isinstance(source, (str, os.PathLike)):
        source_file = fsspec.open(str(source), "rb")
    else:
        source_file = source

    if isinstance(destination, (str, os.PathLike)):
        # adapt the file mode of the destination depending on the source file.
        destination_mode = "wb"
        if hasattr(source_file, "mode"):
            destination_mode = "wb" if "b" in getattr(source_file, "mode") else "w"
        elif isinstance(source_file, io.BytesIO):
            destination_mode = "wb"
        elif isinstance(source_file, io.StringIO):
            destination_mode = "w"

        destination_file = fsspec.open(str(destination), destination_mode)
    else:
        destination_file = destination

    if not is_file(source_file):  # type: ignore
        raise ValueError(f"The file being copied does not exist or is not a file: {source}")

    if not force and is_file(destination_file):  # type: ignore
        raise ValueError(f"The destination file to copy already exists: {destination}")

    with source_file as source_stream:
        with destination_file as destination_stream:
            if chunk_size is None:
                # copy without chunks
                destination_stream.write(source_stream.read())  # type: ignore

            else:
                # copy with chunks

                # determine the size of the source file
                source_size = None
                if progress:
                    source_size = get_size(source)

                pbar = None
                if progress:
                    tqdm = _import_tqdm()

                    if tqdm is None:
                        raise ImportError(
                            "If the progress bar is enabled, you must have `tqdm` "
                            "installed: `conda install tqdm`."
                        )
                    else:
                        # init progress bar
                        pbar = tqdm(
                            total=source_size,
                            leave=leave_progress,
                            disable=not progress,
                            unit="B",
                            unit_divisor=1024,
                            unit_scale=True,
                        )

                # start the loop
                while True:
                    data = source_stream.read(chunk_size)  # type: ignore
                    if not data:
                        break
                    destination_stream.write(data)  # type: ignore

                    if pbar is not None:
                        pbar.update(chunk_size)

                if pbar is not None:
                    pbar.close()


def mkdir(dir_path: Union[str, os.PathLike], exist_ok: bool = False):
    """Create a directory.

    Args:
        dir_path: The path of the directory to create.
        exist_ok: Whether to ignore the error if the directory
            already exists.
    """
    fs = get_mapper(str(dir_path)).fs
    fs.mkdirs(str(dir_path), exist_ok=exist_ok)


def glob(path: str, detail: bool = False, **kwargs) -> List[str]:
    """Find files by glob-matching.

    Args:
        path: A glob-style path.
    """
    # Get the list of paths
    fs = get_mapper(path).fs
    paths = fs.glob(path, detail=detail, **kwargs)
    paths = [fsspec.utils._unstrip_protocol(d, fs) for d in paths]
    return paths


def copy_dir(
    source: Union[str, pathlib.Path],
    destination: Union[str, pathlib.Path],
    force: bool = False,
    progress: bool = False,
    leave_progress: bool = True,
    file_progress: bool = False,
    file_leave_progress: bool = False,
    chunk_size: Optional[int] = None,
):
    """Copy one directory to another location across different filesystem (local, S3, GCS, etc).

    Note that if both FS from source and destination are the same, progress won't be shown.

    Args:
        source: Path to the source directory.
        destination: Path to the destination directory.
        chunk_size: the chunk size to use. If progress is enabled the chunk
            size is `None`, it is set to 2048.
        force: whether to overwrite the destination directory if it exists.
        progress: Whether to display a progress bar.
        leave_progress: Whether to hide the progress bar once the copy is done.
        file_progress: Whether to display a progress bar for each file.
        file_leave_progress: Whether to hide the progress bar once a file copy is done.
        chunk_size: See `po.utils.fs.copy_file`.
    """

    source = str(source)
    destination = str(destination)

    source_fs = get_mapper(source).fs
    destination_fs = get_mapper(destination).fs

    # Sanity check
    if not is_dir(source):
        raise ValueError(f"The directory being copied does not exist or is not a directory: {source}")

    if not force and is_dir(destination):
        raise ValueError(f"The destination folder to copy already exists: {destination}")

    # If both fs are the same then we just rely on the internal `copy` method
    # which is much faster.
    if destination_fs.__class__ == source_fs.__class__:
        source_fs.copy(source, destination, recursive=True)
        return

    # Get all input paths with details
    # NOTE(hadim): we could have use `.glob(..., detail=True)` here but that API is inconsistent
    # between the backends resulting in different object types being returned (dict, list, etc).
    detailed_paths = source_fs.find(source, withdirs=True, detail=True)
    detailed_paths = list(detailed_paths.values())

    # Get list of input types
    input_types = [d["type"] for d in detailed_paths]

    # Get list of input path + add protocol if needed
    input_paths = [d["name"] for d in detailed_paths]
    input_paths = [fsspec.utils._unstrip_protocol(p, source_fs) for p in input_paths]

    # Build all the output paths
    output_paths: List[str] = fsspec.utils.other_paths(input_paths, destination)  # type: ignore

    def _copy_source_to_destination(input_path, input_type, output_path):
        # A directory
        if input_type == "directory":
            destination_fs.mkdir(output_path)

        # A file
        else:
            copy_file(
                input_path,
                output_path,
                force=force,
                progress=file_progress,
                leave_progress=file_leave_progress,
                chunk_size=chunk_size,
            )

    # Copy source files/directories to destination in parallel
    parallelized(
        _copy_source_to_destination,
        inputs_list=list(zip(input_paths, input_types, output_paths)),
        arg_type="args",
        progress=progress,
        tqdm_kwargs=dict(leave=leave_progress),
        scheduler="threads",
    )


def hash_file(file: Union[str, os.PathLike, io.BytesIO, io.IOBase], chunk_size: int = 4096):
    """Return the md5 hash of a file."""

    md5 = hashlib.md5()

    if isinstance(file, (io.BytesIO, io.TextIOBase, io.BufferedIOBase)):
        sentinel = b""
        if isinstance(file, io.TextIOBase):
            sentinel = ""

        for block in iter(lambda: file.read(chunk_size), sentinel):
            if isinstance(block, str):
                block = block.encode()

            md5.update(block)
        file.seek(0)

    elif is_file(file):
        with fsspec.open(file, "rb") as f:
            for block in iter(lambda: f.read(chunk_size), b""):
                md5.update(block)

    return md5.hexdigest()



import functools
from typing import Any, List, Union, Tuple

import pandas as pd

from rdkit.Chem import FindMolChiralCenters

import datamol as dm
from datamol.mol import Mol
from datamol.isomers._enumerate import count_stereoisomers

UNIQUE_ID = "molhash_id"
NO_STEREO_UNIQUE_ID = "molhash_id_no_stereo"
SMILES_COL = "smiles"
STEREO_DEF = "stereo_defined"
NUM_UNDEF_STEREO_CENTER = "num_undefined_stereo_center"
NUM_DEF_STEREO_CENTER = "num_defined_stereo_center"
NUM_STEREO_CENTER = "num_stereo_center"
NUM_STEREOISOMERS = "num_stereoisomers"
NUM_UNDEF_STEREOISOMERS = "num_undefined_stereoisomers"
UNDEF_EZ = "undefined_E/Z"  # E/Z diastereomer
UNDEF_ED = "undefined_E_D"  # enantiomers and diastereomer


def _num_stereo_centers(mol: Mol) -> Tuple[int]:
    """Get the number of defined and undefined stereo centers of a given molecule
        by accessing the all and only defined stereo centers.
        It's to facilitate the analysis of the stereo isomers.
        None will be return if there is no stereo centers in the molecule.

     Args:
         mol: Molecule

    Returns:
        nun_defined_centers: Number of defined stereo centers.
        nun_undefined_centers: Number of undefined stereo centers.

    See Also:
        <rdkit.Chem.FindMolChiralCenters>

    """
    num_all_centers = len(FindMolChiralCenters(mol, force=True, includeUnassigned=True))
    num_defined_centers = len(FindMolChiralCenters(mol, force=True, includeUnassigned=False))
    if num_all_centers == 0:
        return 0, 0, 0
    nun_undefined_centers = num_all_centers - num_defined_centers
    return num_all_centers, num_defined_centers, nun_undefined_centers


def _curate_mol(
    mol: Union[Mol, str],
    remove_salt_solvent: bool = True,
    remove_stereo: bool = False,
) -> dict:
    """Clean and standardize molecule to ensure the quality molecular structures from various resources.
       It comes with the option of remove salts/solvents and stereochemistry information from the molecule.

    Args:
        mol: Molecule
        remove_salt_solvent: When set to 'True', all disconnected salts and solvents
                             will be removed from molecule. In most of the cases,
                             the salts/solvents are recommended to be removed.
        remove_stereo: Whether remove stereochemistry information from molecule.
                       If it's known that the stereochemistry do not contribute to the bioactivity of interest,
                       the stereochemistry information can be removed.

    Returns:
        mol_dict: Dictionary of curated molecule with unique ids.
    """
    with dm.without_rdkit_log():
        mol = dm.to_mol(mol)
        # fix mol
        mol = dm.fix_mol(mol)
        # sanitize molecule
        mol = dm.sanitize_mol(mol, sanifix=True, charge_neutral=False)

        if remove_salt_solvent:
            # standardize here to ensure the success the substructure matching for
            mol = dm.standardize_mol(
                mol=mol,
                disconnect_metals=False,
                reionize=True,
                normalize=True,
                uncharge=False,
                stereo=not remove_stereo,
            )
            # remove salts
            mol = dm.remove_salts_solvents(mol)

        # remove stereochemistry information
        if remove_stereo:
            mol = dm.remove_stereochemistry(mol)

        # standardize
        mol = dm.standardize_mol(
            mol=mol,
            disconnect_metals=False,
            reionize=True,
            normalize=True,
            uncharge=False,
            stereo=not remove_stereo,
        )

        # number of possible stereoisomers
        num_stereoisomers = count_stereoisomers(
            mol=mol, undefined_only=False, rationalise=True, clean_it=True
        )

        # number of undefined stereoisomers
        num_undefined_stereoisomers = count_stereoisomers(
            mol=mol, undefined_only=True, rationalise=True, clean_it=True
        )

        # number of stereocenters
        num_all_centers, num_defined_centers, num_undefined_centers = _num_stereo_centers(mol)

        mol_dict = {
            SMILES_COL: dm.to_smiles(mol, canonical=True),
            UNIQUE_ID: dm.hash_mol(mol),
            NO_STEREO_UNIQUE_ID: dm.hash_mol(mol, hash_scheme="no_stereo"),
            NUM_STEREO_CENTER: num_all_centers,
            NUM_UNDEF_STEREO_CENTER: num_undefined_centers,
            NUM_DEF_STEREO_CENTER: num_defined_centers,
            NUM_STEREOISOMERS: num_stereoisomers,
            NUM_UNDEF_STEREOISOMERS: num_undefined_stereoisomers,
            # None of the stereochemistry is defined in the molecule
            UNDEF_ED: num_defined_centers == 0 and num_all_centers > 0,
            # Undefined EZ stereochemistry which has no stereocenter.
            UNDEF_EZ: num_all_centers == 0 and num_undefined_stereoisomers > 0,
        }
        return mol_dict


def run_chemistry_curation(
    mols: List[Union[Mol, str]],
    ignore_stereo: bool = False,
    remove_salt_solvent: bool = True,
    progress: bool = False,
    **parallelized_args: Any,
) -> pd.DataFrame:
    """Perform curation on the input molecules.

    Args:
        mols: List of molecules.
        ignore_stereo: If set to `True`, the stereochemisty information will be ignored and removed from molecule
                       during the curation procedure. If it's known that the stereochemistry do not contribute to
                       the bioactivity of interest, the stereochemistry information can be ignored.
                       It's recommended to always keep the stererochemistry information.
        remove_salt_solvent: When set to 'True', all disconnected salts and solvents
                             will be removed from molecule. In most of the cases,
                             the salts/solvents are recommended to be removed.
        progress: Whether show progress of the parallelization process.
        parallelized_args: Additional parameters for <datamol.utils.parallelized> to control the parallelization process.

    Returns:
        A dataframe which contains the smiles, unique_id with and without stereo chemistry information of the molecules.

    See Also:
        <datamol.utils.parallelized>
    """
    _curate_fn = functools.partial(
        _curate_mol, remove_stereo=ignore_stereo, remove_salt_solvent=remove_salt_solvent
    )
    mol_list = dm.parallelized(
        fn=_curate_fn,
        inputs_list=mols,
        progress=progress,
        **parallelized_args,
    )

    mol_dataframe = pd.DataFrame(mol_list)
    return mol_dataframe



import json
import os.path
from collections import defaultdict
from hashlib import md5
from typing import Dict, Literal, Optional, Tuple, Union

import fsspec
import numpy as np
import pandas as pd
import zarr
from loguru import logger
from pydantic import (
    Field,
    field_validator,
    model_validator,
)

from polaris._artifact import BaseArtifactModel
from polaris.dataset._column import ColumnAnnotation
from polaris.hub.settings import PolarisHubSettings
from polaris.utils import fs
from polaris.utils.constants import DEFAULT_CACHE_DIR
from polaris.utils.dict2html import dict2html
from polaris.utils.errors import InvalidDatasetError, PolarisChecksumError
from polaris.utils.io import get_zarr_root, robust_copy
from polaris.utils.types import AccessType, HttpUrlString, License

# Constants
_SUPPORTED_TABLE_EXTENSIONS = ["parquet"]
_CACHE_SUBDIR = "datasets"
_INDEX_SEP = "#"
_INDEX_FMT = f"{{path}}{_INDEX_SEP}{{index}}"


class Dataset(BaseArtifactModel):
    """Basic data-model for a Polaris dataset, implemented as a [Pydantic](https://docs.pydantic.dev/latest/) model.

    At its core, a dataset in Polaris is a tabular data structure that stores data-points in a row-wise manner.
    A Dataset can have multiple modalities or targets, can be sparse and can be part of one or multiple
     [`BenchmarkSpecification`][polaris.benchmark.BenchmarkSpecification] objects.

    Info: Pointer columns
        Whereas a `Dataset` contains all information required to construct a dataset, it is not ready yet.
        For complex data, such as images, we support storing the content in external blobs of data.
        In that case, the table contains _pointers_ to these blobs that are dynamically loaded when needed.

    Attributes:
        table: The core data-structure, storing data-points in a row-wise manner. Can be specified as either a
            path to a `.parquet` file or a `pandas.DataFrame`.
        md5sum: The checksum is used to verify the version of the dataset specification. If specified, it will
            raise an error if the specified checksum doesn't match the computed checksum.
        readme: Markdown text that can be used to provide a formatted description of the dataset.
            If using the Polaris Hub, it is worth noting that this field is more easily edited through the Hub UI
            as it provides a rich text editor for writing markdown.
        annotations: Each column _can be_ annotated with a [`ColumnAnnotation`][polaris.dataset.ColumnAnnotation] object.
            Importantly, this is used to annotate whether a column is a pointer column.
        source: The data source, e.g. a DOI, Github repo or URI.
    For additional meta-data attributes, see the [`BaseArtifactModel`][polaris._artifact.BaseArtifactModel] class.

    Raises:
        InvalidDatasetError: If the dataset does not conform to the Pydantic data-model specification.
        PolarisChecksumError: If the specified checksum does not match the computed checksum.
    """

    # Public attributes
    # Data
    table: Union[pd.DataFrame, str]
    md5sum: Optional[str] = None

    # Additional meta-data
    readme: str = ""
    annotations: Dict[str, ColumnAnnotation] = Field(default_factory=dict)
    source: Optional[HttpUrlString] = None
    license: Optional[License] = None

    # Config
    cache_dir: Optional[str] = None  # Where to cache the data to if cache() is called.

    # Private attributes
    _path_to_hash: Dict[str, Dict[str, str]] = defaultdict(dict)
    _has_been_warned: bool = False
    _has_been_cached: bool = False

    @field_validator("table")
    def _validate_table(cls, v):
        """If the table is not a dataframe yet, assume it's a path and try load it."""
        if not isinstance(v, pd.DataFrame):
            if not fs.is_file(v) or fs.get_extension(v) not in _SUPPORTED_TABLE_EXTENSIONS:
                raise InvalidDatasetError(f"{v} is not a valid DataFrame or .parquet path.")
            v = pd.read_parquet(v)
        return v

    @model_validator(mode="after")
    @classmethod
    def _validate_model(cls, m: "Dataset"):
        """If a checksum is provided, verify it matches what the checksum should be.
        If no checksum is provided, make sure it is set.
        If no cache_dir is provided, set it to the default cache dir and make sure it exists
        """

        # Verify that all annotations are for columns that exist
        if any(k not in m.table.columns for k in m.annotations):
            raise InvalidDatasetError("There is annotations for columns that do not exist")

        # Set a default for missing annotations and convert strings to Modality
        for c in m.table.columns:
            if c not in m.annotations:
                m.annotations[c] = ColumnAnnotation()

        # Verify the checksum
        # NOTE (cwognum): Is it still reasonable to always verify this as the dataset size grows?
        actual = m.md5sum
        expected = cls._compute_checksum(m.table)

        if actual is None:
            m.md5sum = expected
        elif actual != expected:
            raise PolarisChecksumError(
                "The dataset md5sum does not match what was specified in the meta-data. "
                f"{actual} != {expected}"
            )

        # Set the default cache dir if none and make sure it exists
        if m.cache_dir is None:
            m.cache_dir = fs.join(DEFAULT_CACHE_DIR, _CACHE_SUBDIR, m.name, m.md5sum)
        fs.mkdir(m.cache_dir, exist_ok=True)

        return m

    @staticmethod
    def _compute_checksum(table):
        """Computes a hash of the dataset.

        This is meant to uniquely identify the dataset and can be used to verify the version.

        1. Is not sensitive to the ordering of the columns or rows in the table.
        2. Purposefully does not include the meta-data (source, description, name, annotations).
        3. For any pointer column, it uses a hash of the path instead of the file contents.
            This is a limitation, but probably a reasonable assumption that helps practicality.
            A big downside is that as the dataset is saved elsewhere, the hash changes.
        """
        hash_fn = md5()

        # Sort the columns s.t. the checksum is not sensitive to the column-ordering
        df = table.copy(deep=True)
        df = df[sorted(df.columns.tolist())]

        # Use the sum of the row-wise hashes s.t. the hash is insensitive to the row-ordering
        table_hash = pd.util.hash_pandas_object(df).sum()
        hash_fn.update(table_hash)

        checksum = hash_fn.hexdigest()
        return checksum

    def get_data(self, row: Union[str, int], col: str) -> np.ndarray:
        """Since the dataset might contain pointers to external files, data retrieval is more complicated
        than just indexing the `table` attribute. This method provides an end-point for seamlessly
        accessing the underlying data.

        Args:
            row: The row index in the `Dataset.table` attribute
            col: The column index in the `Dataset.table` attribute

        Returns:
            A numpy array with the data at the specified indices. If the column is a pointer column,
                the content of the referenced file is loaded to memory.
        """

        def _load(p: str, index: Optional[int]) -> np.ndarray:
            """Tiny helper function to reduce code repetition."""
            arr = zarr.convenience.load(p)
            if index is not None:
                arr = arr[index]
            return arr

        value = self.table.loc[row, col]
        if not self.annotations[col].is_pointer:
            return value

        value, index = self._split_index_from_path(value)

        # In the case it is a pointer column, we need to load additional data into memory
        # We first check if the data has been downloaded to the cache.
        path = self._get_cache_path(column=col, value=value)
        if fs.exists(path):
            return _load(path, index)

        # If it doesn't exist, we load from the original path and warn if not local
        if not fs.is_local_path(value) and not self._has_been_warned:
            logger.warning(
                f"You're loading data from a remote location. "
                f"To speed up this process, consider caching the dataset first "
                f"using {self.__class__.__name__}.cache()"
            )
            self._has_been_warned = True
        return _load(value, index)

    def upload_to_hub(
        self,
        env_file: Optional[Union[str, os.PathLike]] = None,
        settings: Optional[PolarisHubSettings] = None,
        cache_auth_token: bool = True,
        access: Optional[AccessType] = "private",
        **kwargs: dict,
    ):
        """
        Very light, convenient wrapper around the
        [`PolarisHubClient.upload_dataset`][polaris.hub.client.PolarisHubClient.upload_dataset] method.
        """
        from polaris.hub.client import PolarisHubClient

        with PolarisHubClient(
            env_file=env_file, settings=settings, cache_auth_token=cache_auth_token, **kwargs
        ) as client:
            return client.upload_dataset(self, access)

    @classmethod
    def from_zarr(cls, path: str) -> "Dataset":
        """Parse a [.zarr](https://zarr.readthedocs.io/en/stable/index.html) hierarchy into a Polaris `Dataset`.

        In short: A `.zarr` file can contain groups and arrays, where each group can again contain groups and arrays.
        Additional user attributes (for any array or group) are saved as JSON files.

        Within Polaris:

        1. Each subgroup of the root group corresponds to a single column.
        2. Each subgroup can contain:
            - A single array with _all_ datapoints.
            - A single array _per_ datapoint.
        3. Additional meta-data is saved to the user attributes of the root group.
        3. The indices are required to be integers.

        Tip: Tutorial
            To learn more about the zarr format, see the
            [tutorial](../tutorials/dataset_zarr.ipynb).

        Warning: Beta functionality
            This feature is still in beta and the API will likely change. Please report any issues you encounter.

        Args:
            path: The path to the root of the `.zarr` directory. Should be compatible with fsspec.
        """

        logger.warning(
            "We are still testing to save and load from .zarr files. "
            "This part of the API will likely change."
        )

        root = zarr.open(path, "r")

        # Get the user attributes
        attrs = root.attrs.asdict()

        # TODO (cwognum): This is outdated and needs to be updated.
        possible_user_attr = ["name", "description", "source", "annotations"]
        attrs = {k: v for k, v in attrs.items() if k in possible_user_attr}

        # Set the annotations
        attrs["annotations"] = attrs.get("annotations", {})
        for column_label in root.group_keys():
            obj = attrs["annotations"].get(column_label, {})
            obj = ColumnAnnotation.model_validate(obj)
            obj.is_pointer = True
            attrs["annotations"][column_label] = obj

        # Construct the table
        # Parse any group into a column
        data = defaultdict(dict)
        for col, group in root.groups():
            keys = list(group.array_keys())

            if len(keys) == 1:
                arr = group[keys[0]]
                for i, arr_row in enumerate(arr):
                    # In case all data is saved in a single array, we construct a path with an index suffix.
                    data[col][i] = _INDEX_FMT.format(path=fs.join(path, arr.name), index=i)

            else:
                for name, arr in group.arrays():
                    try:
                        name = int(name)
                    except ValueError as error:
                        raise InvalidDatasetError(
                            "All names for arrays in the .zarr archive are required to be integers."
                        ) from error
                    data[col][name] = fs.join(path, arr.path)

        # Construct the dataset
        table = pd.DataFrame(data)
        return cls(table=table, **attrs)

    def to_zarr(
        self,
        destination: str,
        array_mode: Dict[str, Literal["single", "multiple"]],
    ) -> str:
        """Saves a dataset to a .zarr file. For more information on the resulting structure,
        see [`from_zarr`][polaris.dataset.Dataset.from_zarr].

        Tip: Tutorial
            To learn more about the zarr format, see the
            [tutorial](../tutorials/dataset_zarr.ipynb).

        Warning: Beta functionality
            This feature is still in beta and the API will likely change. Please report any issues you encounter.

        Args:
            destination: The _directory_ to save the associated data to.
            array_mode: For each of the columns, whether to save all datapoints in a single array
                or create an array per datapoint. Should be one of "single" or "multiple".

        Returns:
            The path to the root zarr file.
        """

        logger.warning(
            "We are still testing to save and load from .zarr files. "
            "This part of the API will likely change."
        )

        if array_mode not in ["single", "multiple"]:
            raise ValueError(f"array_mode should be one of 'single' or 'multiple', not {array_mode}")

        fs.mkdir(destination, exist_ok=True)
        path = fs.join(destination, "dataset.zarr")

        if not isinstance(array_mode, dict):
            array_mode = {k: array_mode for k in self.table.columns}

        root = zarr.open(path, "w")
        for col in self.table.columns:
            group = root.create_group(col)

            # Load an example to get the dtype and shape
            example = self.get_data(row=0, col=col)

            if array_mode[col] == "single":
                # Create one big array for all datapoints
                shape = (len(self.table), *example.shape)
                arr = group.empty(col, shape=shape, dtype=example.dtype)

                for row in self.table.index:
                    # Save the data to the array
                    arr[row] = self.get_data(row=row, col=col)
            else:
                for row in self.table.index:
                    # Create an array per datapoint
                    group.array(row, self.get_data(row=row, col=col))

        # Save the meta-data
        # TODO (cwognum): This is outdated and needs to be updated.
        root.user_attrs = {
            "name": self.name,
            "description": self.description,
            "source": self.source,
            "annotations": {k: v.model_dump() for k, v in self.annotations.items()},
        }
        return path

    @classmethod
    def from_json(cls, path: str):
        """Loads a benchmark from a JSON file.
        Overrides the method from the base class to remove the caching dir from the file to load from,
        as that should be user dependent.

        Args:
            path: Loads a benchmark specification from a JSON file.
        """
        with fsspec.open(path, "r") as f:
            data = json.load(f)
        data.pop("cache_dir", None)
        return cls.model_validate(data)

    def to_json(self, destination: str) -> str:
        """
        Save the dataset to a destination directory as a JSON file.

        Warning: Multiple files
            Perhaps unintuitive, this method creates multiple files.

            1. `/path/to/destination/dataset.json`: This file can be loaded with
                [`Dataset.from_json`][polaris.dataset.Dataset.from_json].
            2. `/path/to/destination/table.parquet`: The `Dataset.table` attribute is saved here.
            3. _(Optional)_ `/path/to/destination/data/*`: Any additional blobs of data referenced by the
                    pointer columns will be stored here.

        Args:
            destination: The _directory_ to save the associated data to.

        Returns:
            The path to the JSON file.
        """
        fs.mkdir(destination, exist_ok=True)
        table_path = fs.join(destination, "table.parquet")
        dataset_path = fs.join(destination, "dataset.json")
        pointer_dir = fs.join(destination, "data")

        # Save additional data
        new_table = self._copy_and_update_pointers(pointer_dir, inplace=False)

        # Lu: Avoid serilizing and sending None to hub app.
        serialized = self.model_dump(exclude={"cache_dir"}, exclude_none=True)
        serialized["table"] = table_path

        # We need to recompute the checksum, as the pointer paths have changed
        serialized["md5sum"] = self._compute_checksum(new_table)

        new_table.to_parquet(table_path)
        with fsspec.open(dataset_path, "w") as f:
            json.dump(serialized, f)

        return dataset_path

    def cache(self, cache_dir: Optional[str] = None) -> str:
        """Caches the dataset by downloading all additional data for pointer columns to a local directory.

        Args:
            cache_dir: The directory to cache the data to. If not provided,
                this will fall back to the `Dataset.cache_dir` attribute

        Returns:
            The path to the cache directory.
        """

        if cache_dir is not None:
            self.cache_dir = cache_dir

        if not self._has_been_cached:
            self._copy_and_update_pointers(self.cache_dir, inplace=True)
            self._has_been_cached = True
        return self.cache_dir

    def _get_cache_path(self, column: str, value: str) -> Optional[str]:
        """
        Returns where the data _would be_ cached for any entry in the pointer columns,
        or None if the column is not a pointer column.
        """
        if not self.annotations[column].is_pointer:
            return

        if value not in self._path_to_hash[column]:
            h = md5(value.encode("utf-8")).hexdigest()

            value, _ = self._split_index_from_path(value)
            ext = fs.get_extension(value)
            dst = fs.join(self.cache_dir, column, f"{h}.{ext}")

            # The reason for caching the path is to speed-up retrieval. Hashing can be slow and with large
            # datasets this could become a bottleneck.
            self._path_to_hash[column][value] = dst

        return self._path_to_hash[column][value]

    def size(self):
        return len(self), len(self.table.columns)

    def _split_index_from_path(self, path: str) -> Tuple[str, Optional[int]]:
        """
        Paths can have an additional index appended to them.
        This extracts that index from the path.
        """
        index = None
        if _INDEX_SEP in path:
            path, index = path.split(_INDEX_SEP)
            index = int(index)
        return path, index

    def _copy_and_update_pointers(
        self, save_dir: str, table: Optional[pd.DataFrame] = None, inplace: bool = False
    ) -> pd.DataFrame:
        """Copy and update the path in the table to the new destination"""

        def fn(path):
            """Helper function that can be used within Pandas apply to copy and update all files"""

            # We copy the entire .zarr hierarchy
            root = get_zarr_root(path)
            if root is None:
                raise NotImplementedError(
                    "Only the .zarr file format is currently supported for pointer columns"
                )

            # We could introduce name collisions here and thus use a hash of the original path for the destination
            dst = fs.join(save_dir, f"{md5(root.encode('utf-8')).hexdigest()}.zarr")
            robust_copy(root, dst)

            diff = os.path.relpath(path, root)
            dst = fs.join(dst, diff)
            return dst

        if table is None:
            table = self.table
        if not inplace:
            table = self.table.copy(deep=True)

        for c in table.columns:
            if self.annotations[c].is_pointer:
                table[c] = table[c].apply(fn)
        return table

    def _repr_dict_(self) -> dict:
        """Utility function for pretty-printing to the command line and jupyter notebooks"""
        repr_dict = self.model_dump()
        repr_dict.pop("table")
        return repr_dict

    def _repr_html_(self):
        """For pretty-printing in Jupyter Notebooks"""
        return dict2html(self._repr_dict_())

    def __len__(self):
        return len(self.table)

    def __repr__(self):
        return json.dumps(self._repr_dict_(), indent=2)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        """Whether two datasets are equal is solely determined by the checksum"""
        if not isinstance(other, Dataset):
            return False
        return self.md5sum == other.md5sum



import json
import os
from datetime import datetime
from typing import ClassVar, Optional, Union

import pandas as pd
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PrivateAttr,
    computed_field,
    field_serializer,
    field_validator,
)
from pydantic.alias_generators import to_camel

from polaris._artifact import BaseArtifactModel
from polaris.evaluate import Metric
from polaris.hub.settings import PolarisHubSettings
from polaris.utils.dict2html import dict2html
from polaris.utils.errors import InvalidResultError
from polaris.utils.misc import sluggify
from polaris.utils.types import AccessType, HttpUrlString, HubOwner, HubUser, SlugCompatibleStringType

# Define some helpful type aliases
TestLabelType = str
TargetLabelType = str


class ResultRecords(BaseModel):
    """
    A grouped, tabular data-structure to save the actual data of the benchmark results.
    The grouping by test set and target label is done to make it easier to build leaderboards.
    """

    test_set: TestLabelType
    target_label: TargetLabelType
    scores: dict[Union[Metric, str], float]

    # Model config
    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True)

    @field_validator("scores")
    def validate_scores(cls, v):
        validated = {}
        for metric, score in v.items():
            if not isinstance(metric, Metric):
                try:
                    metric = Metric[metric]
                except KeyError as error:
                    raise ValueError("Invalid metric name") from error
            validated[metric] = score
        return validated

    @field_serializer("scores")
    def serialize_scores(self, value: dict):
        """Change from the Metric enum to a string representation"""
        return {metric.name: score for metric, score in value.items()}


ResultsType = Union[pd.DataFrame, list[Union[ResultRecords, dict]]]


class BenchmarkResults(BaseArtifactModel):
    """Class for saving benchmarking results

    This object is returned by [`BenchmarkSpecification.evaluate`][polaris.benchmark.BenchmarkSpecification.evaluate].
    In addition to the metrics on the test set, it contains additional meta-data and logic to integrate
    the results with the Polaris Hub.

    The actual results are saved in the `results` field using the following tabular format:

    | Test set | Target label | Metric | Score |
    | -------- | ------------ | ------ | ----- |
    | test_iid | EGFR_WT      | AUC    | 0.9   |
    | test_ood | EGFR_WT      | AUC    | 0.75  |
    |    ...   |      ...     |  ...   |  ...  |
    | test_ood | EGFR_L858R   | AUC    | 0.79  |

    question: Categorizing methods
        An open question is how to best categorize a methodology (e.g. a model).
        This is needed since we would like to be able to aggregate results across benchmarks too,
        to say something about which (type of) methods performs best _in general_.

    Attributes:
        results: Benchmark results are stored directly in a dataframe or in a serialized, JSON compatible dict
            that can be decoded into the associated tabular format.
        benchmark_name: The name of the benchmark for which these results were generated.
            Together with the benchmark owner, this uniquely identifies the benchmark on the Hub.
        benchmark_owner: The owner of the benchmark for which these results were generated.
            Together with the benchmark name, this uniquely identifies the benchmark on the Hub.
        github_url: The URL to the GitHub repository of the code used to generate these results.
        paper_url: The URL to the paper describing the methodology used to generate these results.
        contributors: The users that are credited for these results.
        _created_at: The time-stamp at which the results were created. Automatically set.
    For additional meta-data attributes, see the [`BaseArtifactModel`][polaris._artifact.BaseArtifactModel] class.
    """

    # Define the columns of the results table
    RESULTS_COLUMNS: ClassVar[list[str]] = ["Test set", "Target label", "Metric", "Score"]

    # Data
    results: ResultsType
    benchmark_name: SlugCompatibleStringType = Field(..., frozen=True)
    benchmark_owner: Optional[HubOwner] = Field(None, frozen=True)

    # Additional meta-data
    github_url: Optional[HttpUrlString] = None
    paper_url: Optional[HttpUrlString] = None
    contributors: Optional[list[HubUser]] = None

    # Private attributes
    _created_at: datetime = PrivateAttr(default_factory=datetime.now)

    @computed_field
    @property
    def benchmark_artifact_id(self) -> str:
        return f"{self.benchmark_owner}/{sluggify(self.benchmark_name)}"

    @field_validator("results")
    def _validate_results(cls, v):
        """Ensure the results are a valid dataframe and have the expected columns"""

        # If not a dataframe, assume it is a JSON-serialized export of a dataframe.
        if not isinstance(v, pd.DataFrame):
            try:
                df = pd.DataFrame(columns=cls.RESULTS_COLUMNS)
                for record in v:
                    if isinstance(record, dict):
                        record = ResultRecords(**record)

                    for metric, score in record.scores.items():
                        df.loc[len(df)] = {
                            "Test set": record.test_set,
                            "Target label": record.target_label,
                            "Metric": metric,
                            "Score": score,
                        }
                v = df

            except (ValueError, UnicodeDecodeError) as error:
                raise InvalidResultError(
                    f"The provided dictionary cannot be parsed into a {cls.__name__} instance."
                ) from error

        # Check if the dataframe contains _only_ the expected columns
        if set(v.columns) != set(cls.RESULTS_COLUMNS):
            raise InvalidResultError(
                f"The results dataframe should have the following columns: {cls.RESULTS_COLUMNS}"
            )

        # Check if the results are not empty
        if v.empty:
            raise InvalidResultError("The results dataframe is empty")

        # NOTE (cwognum): Since we have a reference to the benchmark, I considered validating the values in the
        #  columns as well (e.g. are all metrics, targets and test sets actually part of the benchmark).
        #  However, to keep this class light-weight, I did not want to add a strict dependency on the full benchmark class.
        #  Especially because validation will happen on the Hub as well before it is shown there.
        return v

    @field_serializer("results")
    def _serialize_results(self, value: ResultsType):
        """Change from the Metric enum to a string representation"""
        self.results["Metric"] = self.results["Metric"].apply(
            lambda x: x.name if isinstance(x, Metric) else x
        )

        serialized = []
        grouped = self.results.groupby(["Test set", "Target label"])
        for (test_set, target_label), group in grouped:
            metrics = {row["Metric"]: row["Score"] for _, row in group.iterrows()}
            record = ResultRecords(test_set=test_set, target_label=target_label, scores=metrics)
            serialized.append(record.model_dump(by_alias=True))

        return serialized

    def upload_to_hub(
        self,
        env_file: Optional[Union[str, os.PathLike]] = None,
        settings: Optional[PolarisHubSettings] = None,
        cache_auth_token: bool = True,
        access: Optional[AccessType] = "private",
        **kwargs: dict,
    ):
        """
        Very light, convenient wrapper around the
        [`PolarisHubClient.upload_results`][polaris.hub.client.PolarisHubClient.upload_results] method.
        """
        from polaris.hub.client import PolarisHubClient

        with PolarisHubClient(
            env_file=env_file, settings=settings, cache_auth_token=cache_auth_token, **kwargs
        ) as client:
            return client.upload_results(self, access)

    def _repr_dict_(self) -> dict:
        """Utility function for pretty-printing to the command line and jupyter notebooks"""
        repr_dict = self.model_dump(exclude=["results"])

        df = self.results.copy(deep=True)
        df["Metric"] = df["Metric"].apply(lambda x: x.name if isinstance(x, Metric) else x)
        repr_dict["results"] = json.loads(df.to_json(orient="records"))

        return repr_dict

    def _repr_html_(self):
        """For pretty-printing in Jupyter Notebooks"""
        return dict2html(self._repr_dict_())

    def __len__(self):
        return len(self.table)

    def __repr__(self):
        return json.dumps(self._repr_dict_(), indent=2)



from enum import Enum
from typing import List, TypeVar, Union
import numpy as np
from pydantic import BaseModel, Field
from scipy import stats

from sklearn.base import (
    BaseEstimator,
    OneToOneFeatureMixin,
    TransformerMixin,
    _fit_context,
)
from sklearn.covariance import EllipticEnvelope
from sklearn.ensemble import IsolationForest
from sklearn.neighbors import LocalOutlierFactor
from sklearn.svm import OneClassSVM
from sklearn.utils import check_array
from sklearn.utils.validation import check_is_fitted


PandasDataFrame = TypeVar("pandas.core.frame.DataFrame")
NumpyNDArray = TypeVar("numpy.ndarray")


class LabelOrder(Enum):
    acs = "ascending"
    desc = "descending"


def discretizer(
    X: np.ndarray,
    thresholds: Union[np.ndarray, list],
    copy: bool = True,
    allow_nan: bool = True,
    label_order: LabelOrder = LabelOrder.acs.value,
):
    """Thresholding of array-like or scipy.sparse matrix into binary or multiclass labels.

    Args:
        X : The data to discretize, element by element.
            scipy.sparse matrices should be in CSR or CSC format to avoid an
            un-necessary copy.

        thresholds: Feature values below or equal to this are replaced by 0, above it by 1.
            Threshold may not be less than 0 for operations on sparse matrices.

        copy: Set to False to perform inplace discretization and avoid a copy
            (if the input is already a numpy array or a scipy.sparse CSR / CSC
            matrix and if axis is 1).

        allow_nan: Set to True to allow nans in the array for discretization. Otherwise,
            an error will be raised instead.

        label_order: The continuous values are discretized to labels 0, 1, 2, .., N with respect to given
            threshold bins [threshold_1, threshold_2,.., threshould_n].
            When set to 'ascending', the class label is in ascending order with the threshold
            bins that `0` represents negative class or lower class, while 1, 2, 3 are for higher classes.
            When set to 'descending' the class label is in ascending order with the threshold bins.
            Sometimes the positive labels are on the left side of provided threshold.
            E.g. For binarization with threshold [0.5],  the positive label is defined
            by`X < 0.5`. In this case, `label_order` should be `descending`.

    Returns:
        X_tr: The transformed data.

    See Also:
        Discretizer : Performs multiclass discretizer using the Transformer API
    """
    if label_order not in [LabelOrder.acs.value, LabelOrder.desc.value]:
        raise ValueError(
            f"Please specify `label_order` using the {LabelOrder.desc.value} or {LabelOrder.acs.value}"
        )

    X = check_array(
        X,
        accept_sparse=["csr", "csc"],
        copy=copy,
        force_all_finite="allow-nan" if allow_nan else True,
        ensure_2d=False,
    )

    nan_idx = np.isnan(X)

    binarize = True if len(thresholds) == 1 else False
    if label_order == LabelOrder.desc.value:
        thresholds = np.flip(thresholds)
    X = np.digitize(X, thresholds)
    if allow_nan:
        X = X.astype(np.float64)
        X[nan_idx] = np.nan
    if binarize and label_order == LabelOrder.desc.value:
        X = 1 - X
    return X


class Discretizer(OneToOneFeatureMixin, TransformerMixin, BaseEstimator, BaseModel):
    """Discretizer continuous data according to a list thresholds e.g. [threshold_1, threshold_2].

       In the above example, values falls in left of threshold_1 are mapped to 0, values fall in
       the interval between threshold_1 and threshold_2 are mapped to 1, while values fall on the right
       threshold_2 are mapped to 2. The threshold values must be in ascending or descending order.
       To reverse the label order, set `label_order` to `descending`.

    Args:
        thresholds : np.ndarray, default=0.0
            Feature values below or equal to this are replaced by 0, above it by 1.
            Threshold may not be less than 0 for operations on sparse matrices.

        copy : bool, default=True
            Set to False to perform inplace discretization and avoid a copy (if
            the input is already a numpy array or a scipy.sparse CSR matrix).

    See Also:
        discretizer : Equivalent function without the estimator API.
    """

    thresholds: List = Field(default=[0])
    copy_object: bool = Field(default=True, alias="copy")
    label_order: LabelOrder = Field(default=LabelOrder.acs)

    @_fit_context(prefer_skip_nested_validation=True)
    def fit(self, X, y=None):
        """Only validates estimator's parameters.

        This method allows to: (i) validate the estimator's parameters and
        (ii) be consistent with the scikit-learn transformer API.

        Args:
            X : {array-like, sparse matrix} of shape (n_samples, n_features)
                The data.

            y : None
                Ignored.

        Returns:
            self : object
                Fitted transformer.
        """
        self._validate_data(X, accept_sparse="csr")
        return self

    def transform(self, X, copy=None):
        """Convert each element of X to multiclass label.

        Args:
            X : {array-like, sparse matrix} of shape (n_samples, n_features)
                The data to binarize, element by element.
                scipy.sparse matrices should be in CSR format to avoid an
                un-necessary copy.

            copy : bool
                Copy the input X or not.

        Returns
            X_tr : {ndarray, sparse matrix} of shape (n_samples, n_features)
                Transformed array.
        """
        copy = copy if copy is not None else self.copy
        # check_array
        X = self._validate_data(X, accept_sparse=["csr", "csc"], copy=copy, reset=False)
        return discretizer(X, thresholds=self.thresholds, copy=False, label_order=self.label_order)

    def _more_tags(self):
        return {"stateless": True}


def modified_zscore(data: np.ndarray, consistency_correction: float = 1.4826):
    """
    The modified z score is calculated from the median absolute deviation (MAD).
    These values must be multiplied by a constant to approximate the standard deviation.

    The modified z score might be more robust than the standard z score because it relies
    on the median (MED) for calculating the z score.

            modified Z score = (X-MED)/(consistency_correction*MAD)

    """
    median = np.nanmedian(data)

    deviation_from_med = np.array(data) - median

    mad = np.nanmedian(np.abs(deviation_from_med))
    mod_zscore = deviation_from_med / (consistency_correction * mad)
    return mod_zscore, mad


class ZscoreOutlier(BaseModel):
    """Detect outliers by the absolute value of modified Zscore.
    Args:
        threshold: zscore threshold to define the outliers. By default, the outlier is defined if the absolute value
        of modified zscore greater than 3.
        modified: When set to 'True', modified Zscore is used.
    """

    threshold: float = 3
    modified: bool = False
    zscore_: NumpyNDArray = None
    mad: float = None

    def fit(self, X):
        if self.modified:
            self.zscore_, self.mad = modified_zscore(X)
        else:
            self.zscore_ = stats.zscore(X)

    def predict(self, X, y=None):
        check_is_fitted(self)
        result = np.ones_like(X)
        ind = np.argwhere(np.absolute(self.zscore_) > self.threshold)
        result[ind] = -1
        return result

    def fit_predict(self, X, y=None):
        self.fit(X)
        return self.predict(X)


OUTLIER_METHOD = {
    "iso": IsolationForest,
    "lof": LocalOutlierFactor,
    "svm": OneClassSVM,
    "ee": EllipticEnvelope,
    "zscore": ZscoreOutlier,
}


def outlier_detection(X: np.ndarray, method="zscore", **kwargs) -> np.ndarray:
    """Automatic detection of data outliers.

    Args:
        X: Array of values to be analyzed.
        method: Name of the algorithm to be used for outlier detection.
                The supported methods are
                    'iso' for IsolationForest,
                    'lof' for LocalOutlierFactor,
                    'svm' for OneClassSVM,
                    'ee' for EllipticEnvelope, and
                    'zscore' for ZscoreOutlier.
        kwargs: Additional parameters for the outlier detection algorithm.

    Returns:
        outlier_index: Index of detected outliers in the input data X.
    """
    if method not in OUTLIER_METHOD:
        raise ValueError("The detection name must be in 'iso', 'lof', 'svm', 'ee' and 'zscore'.")
    detector = OUTLIER_METHOD.get(method)(**kwargs)
    pred = detector.fit_predict(X)
    outlier_index = np.argwhere(pred == -1)
    return outlier_index



import os.path
import uuid
from typing import Optional

import filelock
import fsspec
from loguru import logger
from tenacity import Retrying
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_fixed

from polaris.utils import fs
from polaris.utils.constants import DEFAULT_CACHE_DIR
from polaris.utils.errors import PolarisChecksumError


def create_filelock(lock_name: str, cache_dir_path: str = DEFAULT_CACHE_DIR):
    """Create an empty lock file into `cache_dir_path/locks/lock_name`"""
    lock_path = fs.join(cache_dir_path, "_lock_files", lock_name)
    with fsspec.open(lock_path, "w", auto_mkdir=True):
        pass
    return filelock.FileLock(lock_path)


def robust_copy(
    source_path: str,
    destination_path: str,
    md5sum: Optional[str] = None,
    max_retries: int = 5,
    wait_after_try: int = 2,
    progress: bool = True,
    leave_progress: bool = True,
    chunk_size: int = 2048,
):
    if not fs.is_file(source_path) and get_zarr_root(source_path) is None:
        raise ValueError(f"{source_path} is a directory and not part of a .zarr hierarchy!")

    if md5sum is None and fs.is_file(source_path):
        # NOTE (cwognum): This effectively means we will not check the checksum of .zarr files.
        #  The reason being that I'm not sure how to effectively compute a checksum for a .zarr
        md5sum = fs.hash_file(source_path)

    artifact_cache_lock = create_filelock(f"artifact_version_{md5sum or uuid.uuid4()}.lock")

    def log_failure(retry_state):
        logger.warning(
            f"""Downloading the artifact from {source_path} to {destination_path} failed. """
            f"""Retrying attempt {retry_state.attempt_number}/{max_retries} """
            f"""after a sleeping period of {wait_after_try} seconds."""
        )

    # This context manager will lock any process that try to download the same file. Only one process
    # will be able to download the artifact and all the other ones will be waiting at that line.
    # Once the lock is released the other processes will call `download_with_checksum` but the download will
    # not happen since the artifact file will already exist and its checksum will be correct.
    with artifact_cache_lock:
        # This loop will retry downloading the artifact for multiple attempts. Downloading an artifact
        # might fail for multiple reasons such as disk IO failures or network failures. The checksum logic
        # and the retry mechanism together allow to be resilient in case of intermitent failures.
        for attempt in Retrying(
            reraise=True,
            stop=stop_after_attempt(max_retries),
            after=log_failure,
            wait=wait_fixed(wait_after_try),
        ):
            with attempt:
                # The checksum logic will only validate an artifact download if its checksum matches
                # the excepted one. If not then it will be deleted and the download will happen again
                # until it succeeds (or until the number of attemps have been reached).
                download_with_checksum(
                    source_path=source_path,
                    destination_path=destination_path,
                    md5sum=md5sum,
                    progress=progress,
                    leave_progress=leave_progress,
                    chunk_size=chunk_size,
                )

    return destination_path


def download_with_checksum(
    source_path: str,
    destination_path: str,
    md5sum: Optional[str],
    progress: bool = False,
    leave_progress: bool = True,
    chunk_size: int = 2048,
):
    """Download an artifact from the bucket to a cache path while checking for its md5sum given a true md5sum.

    Args:
        source_path: The path to the artifact in the bucket.
        destination_path: The path of the artifact in the local cache.
        md5sum: The true md5sum to check against. If None, no checksum is performed but a warning is logged.
        progress: whether to display a progress bar.
        leave_progress: whether to hide the progress bar once the copy is done.
        chunk_size: the chunk size for the download.
    """

    # Download the artifact if not already in the cache.
    if not fs.exists(destination_path):
        if fs.is_dir(source_path):
            fs.copy_dir(
                source_path,
                destination_path,
                progress=progress,
                leave_progress=leave_progress,
                chunk_size=chunk_size,
            )

        else:
            fs.copy_file(
                source_path,
                destination_path,
                progress=progress,
                leave_progress=leave_progress,
                chunk_size=chunk_size,
            )

    # Check the cached artifact has the correct md5sum
    if md5sum is not None:
        cache_md5sum = fs.hash_file(destination_path)
        if cache_md5sum != md5sum:
            file_system = fs.get_mapper(destination_path).fs
            file_system.delete(destination_path)

            raise PolarisChecksumError(
                f"""The destination artifact at {destination_path} has a different md5sum ({cache_md5sum})"""
                f"""than the expected artifact md5sum ({md5sum}). The destination artifact has been deleted. """
            )


def get_zarr_root(path):
    """
    Recursive function to find the root of a .zarr file.
    Finds the highest level directory that has the .zarr extension.
    """
    if os.path.dirname(path) == path:
        # We reached the root of the filesystem
        return
    root = get_zarr_root(os.path.dirname(path))
    if root is None and fs.get_extension(path) == "zarr":
        root = path
    return root



import pytest

from polaris._artifact import BaseArtifactModel
from polaris.utils.types import HubOwner, License


def test_slug_string_type():
    """
    Verifies that the slug is validated correctly.
    Fails if:
    - Is too short (<4 characters)
    - Is too long (>64 characters)
    - Contains something other than lowercase letters, numbers, and hyphens.
    """
    for name in ["", "x", "xx", "xxx", "x" * 65, "invalid@", "invalid!", "InvalidName1", "invalid_name"]:
        with pytest.raises(ValueError):
            HubOwner(slug=name)

    for name in ["valid", "valid-name-1", "x" * 64, "x" * 4]:
        HubOwner(slug=name)


def test_slug_compatible_string_type():
    """Verifies that the artifact name is validated correctly."""

    # Fails if:
    # - Is too short (<4 characters)
    # - Is too long (>64 characters)
    # - Contains non-alphanumeric characters
    for name in ["", "x", "xx", "xxx", "x" * 65, "invalid@", "invalid!"]:
        with pytest.raises(ValueError):
            BaseArtifactModel(name=name)

    # Does not fail
    for name in ["valid", "valid-name", "valid_name", "ValidName1", "Valid_", "Valid-", "x" * 64, "x" * 4]:
        BaseArtifactModel(name=name)


def test_license():
    for license in ["0BSD", "CC-BY-NC-4.0", "MIT"]:
        # If a valid SPDX license, the reference is automatically set
        assert License(id=license).reference is not None

    # If not a valid SPDX license, you must specify a valid reference
    with pytest.raises(ValueError):
        License(id="invalid")
    with pytest.raises(ValueError):
        License(id="invalid", reference="invalid")

    # If you specify a URL, we trust the user that this is a valid license
    License(id="invalid", reference="https://example.com")



import numpy as np
import pandas as pd
import pytest
import zarr
from pydantic import ValidationError

from polaris.dataset import Dataset
from polaris.loader import load_dataset
from polaris.utils import fs
from polaris.utils.errors import PolarisChecksumError


def _equality_test(dataset_1, dataset_2):
    """
    Utility function.

    When saving a dataset to a different location, it should be considered the same dataset
    but currently the dataset checksum is used for equality and with pointer columns,
    the checksum uses the file path, not the file content (which thus changes when saving).

    See also: https://github.com/polaris-hub/polaris/issues/16
    """
    if dataset_1 == dataset_2:
        return True
    if len(dataset_1) != len(dataset_2):
        return False
    if (dataset_1.table.columns != dataset_2.table.columns).all():
        return False

    for i in range(len(dataset_1)):
        for col in dataset_1.table.columns:
            if (dataset_1.get_data(row=i, col=col) != dataset_2.get_data(row=i, col=col)).all():
                return False
    return True


def test_load_data(tmp_path):
    """Test accessing the data, specifically whether pointer columns are properly handled."""

    # Dummy data (could e.g. be a 3D structure or Image)
    arr = np.random.random((100, 100))

    tmpdir = str(tmp_path)
    path = fs.join(tmpdir, "data.zarr")
    zarr.save(path, arr)

    table = pd.DataFrame({"A": [path]}, index=[0])
    dataset = Dataset(table=table, cache_dir=tmpdir, annotations={"A": {"is_pointer": True}})

    # Without caching
    data = dataset.get_data(row=0, col="A")
    assert (data == arr).all()

    # With caching
    dataset.cache()
    data = dataset.get_data(row=0, col="A")
    assert (data == arr).all()


def test_dataset_checksum(test_dataset):
    """Test whether the checksum is a good indicator of whether the dataset has changed in a meaningful way."""

    original = test_dataset.md5sum
    assert original is not None

    # Without any changes, same hash
    kwargs = test_dataset.model_dump()
    Dataset(**kwargs)

    # With unimportant changes, same hash
    kwargs["name"] = "changed"
    kwargs["description"] = "changed"
    kwargs["source"] = "https://changed.com"
    Dataset(**kwargs)

    # Check sensitivity to the row and column ordering
    kwargs["table"] = kwargs["table"].iloc[::-1]
    kwargs["table"] = kwargs["table"][kwargs["table"].columns[::-1]]
    Dataset(**kwargs)

    def _check_for_failure(_kwargs):
        with pytest.raises(ValidationError) as error:
            Dataset(**_kwargs)
            assert error.error_count() == 1  # noqa
            assert isinstance(error.errors()[0], PolarisChecksumError)  # noqa

    # Without any changes, but different hash
    kwargs["md5sum"] = "invalid"
    _check_for_failure(kwargs)

    # With changes, but same hash
    kwargs["md5sum"] = original
    kwargs["table"] = kwargs["table"].iloc[:-1]
    _check_for_failure(kwargs)

    # With changes, but no hash
    kwargs["md5sum"] = None
    dataset = Dataset(**kwargs)
    assert dataset.md5sum is not None


@pytest.mark.parametrize("array_per_datapoint", [True, False])
def test_dataset_from_zarr(
    test_zarr_archive_single_array, test_zarr_archive_multiple_arrays, array_per_datapoint
):
    """Test whether loading works when the zarr archive contains a single array or multiple arrays."""
    archive = test_zarr_archive_multiple_arrays if array_per_datapoint else test_zarr_archive_single_array
    dataset = Dataset.from_zarr(archive)
    assert len(dataset.table) == 100
    for i in range(100):
        assert dataset.get_data(row=i, col="A").shape == (2048,)
        assert dataset.get_data(row=i, col="B").shape == (2048,)


def test_dataset_from_zarr_equality(test_zarr_archive_single_array, test_zarr_archive_multiple_arrays):
    """
    Test whether two methods for specifying .zarr datasets lead to the same dataset.
    This specifically tests whether indexing a single arrow with our custom path syntax works.
    """
    dataset_1 = Dataset.from_zarr(test_zarr_archive_single_array)
    dataset_2 = Dataset.from_zarr(test_zarr_archive_multiple_arrays)
    assert _equality_test(dataset_1, dataset_2)


def test_dataset_from_json(test_dataset, tmpdir):
    """Test whether the dataset can be saved and loaded from json."""
    test_dataset.to_json(str(tmpdir))

    path = fs.join(str(tmpdir), "dataset.json")

    new_dataset = Dataset.from_json(path)
    assert _equality_test(test_dataset, new_dataset)

    new_dataset = load_dataset(path)
    assert _equality_test(test_dataset, new_dataset)


@pytest.mark.parametrize("array_per_datapoint", [True, False])
def test_dataset_from_zarr_to_json_and_back(
    test_zarr_archive_single_array,
    test_zarr_archive_multiple_arrays,
    array_per_datapoint,
    tmpdir,
):
    """
    Test whether a dataset with pointer columns, instantiated from a zarr archive,
    can be saved to and loaded from json.
    """

    tmpdir = str(tmpdir)
    json_dir = fs.join(tmpdir, "json")
    zarr_dir = fs.join(tmpdir, "zarr")

    archive = test_zarr_archive_multiple_arrays if array_per_datapoint else test_zarr_archive_single_array
    dataset = Dataset.from_zarr(archive)
    path = dataset.to_json(json_dir)

    new_dataset = Dataset.from_json(path)
    assert _equality_test(dataset, new_dataset)

    new_dataset = load_dataset(path)
    assert _equality_test(dataset, new_dataset)

    path = new_dataset.to_zarr(zarr_dir, "multiple" if array_per_datapoint else "single")
    new_dataset = load_dataset(path)
    assert _equality_test(dataset, new_dataset)


@pytest.mark.parametrize("array_per_datapoint", [True, False])
def test_dataset_caching(
    test_zarr_archive_single_array,
    test_zarr_archive_multiple_arrays,
    array_per_datapoint,
    tmpdir,
):
    """Test whether the dataset remains the same after caching."""
    archive = test_zarr_archive_multiple_arrays if array_per_datapoint else test_zarr_archive_single_array

    original_dataset = Dataset.from_zarr(archive)
    cached_dataset = Dataset.from_zarr(archive)
    assert original_dataset == cached_dataset

    for i in range(len(cached_dataset)):
        assert not cached_dataset.table.loc[i, "A"].startswith(original_dataset.cache_dir)
        assert not cached_dataset.table.loc[i, "B"].startswith(original_dataset.cache_dir)

    cached_dataset.cache()
    for i in range(len(cached_dataset)):
        assert cached_dataset.table.loc[i, "A"].startswith(original_dataset.cache_dir)
        assert cached_dataset.table.loc[i, "B"].startswith(original_dataset.cache_dir)

    assert _equality_test(cached_dataset, original_dataset)
